/** @file
    Block collect process.
    Collects the decoded blocks and forms a macroblock
    (as a group of 6, 8 or 12 blocks), which is sent to
    the following process.
*/

#include <stdio.h>
#include <string.h>

#include "collect_blocks.h"

#ifdef DEBUG
    extern FILE *logfile;
#endif

/**
    Process initialization function.
    Initializes local variables.
*/
void collect_blocks_init(DOLProcess *p)
{
    sprintf(p->local->name, "coll blocks");
    p->local->in_port_count = 0;
#ifdef DEBUG
    p->local->block_count = 0;
    p->local->mb_count = 0;
#endif

}


/**
    Process fire function, called by the scheduler.
    Reads all decoded blocks of the current macroblock, stores the data
    in one buffer, which is sent to the following process. In the case
    if a skipped macroblock, only the header information of one block is
    read and forwarded to the following process (this is necessary for the
    following process to receive the motion vectors, which are used to build
    the prediction of the skipped macroblock).
*/
int collect_blocks_fire(DOLProcess *p)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
    BlockDispatchedDatatype *bdd = &(p->local->block_dispatched_datatype);
    int block_number;
    char control_code;


    CREATEPORTVAR(port);

    /* start with first block from macroblock */
    block_number = 0;

#ifdef DEBUG
    p->local->mb_count++;
#endif

    /* loop through all blocks of current macroblock */
    do {

        /* build output port name */
        CREATEPORT( port , PORT_IN, 1 , p->local->in_port_count , NUM_OF_BLOCK_PORTS);

        /* read control code */
        DOL_read((void*)port, &control_code, sizeof(char), p);

        if(control_code == CTRL_END_OF_SEQ) {

            collect_blocks_finish(p);
            return 0;

        }

        if((control_code == CTRL_END_OF_GOP) ||
           (control_code == CTRL_END_OF_ATT_GOP)){

            /* pass end_of_gop control code through */
            DOL_write((void*)PORT_OUT, &control_code, sizeof(char), p);

            /* (circularly) increment input port number. this must be done
               according to the dispatch_blocks process, which increments the
               port number after an end_of_gop control code! */
            if(p->local->in_port_count < (NUM_OF_BLOCK_PORTS-1))
                p->local->in_port_count++;
            else
                p->local->in_port_count = 0;

            return 0;
        }

        if(control_code != CTRL_NEXT_BLOCK) {
            Error(p, "CTRL_NEXT_BLOCK expected");
        }

        /* (circularly) increment input port number (it not used until the
           next sprintf call, but must be incremented before the break in case
           of a skipped mb!) */
        if(p->local->in_port_count < (NUM_OF_BLOCK_PORTS-1))
            p->local->in_port_count++;
        else
            p->local->in_port_count = 0;

        /* read gop dispatched datatype */
        DOL_read((void*)port, gdd, sizeof(GopDispatchedDatatype), p);
        DOL_read((void*)port, mdd, sizeof(MbDispatchedDatatype), p); // read mb dispatched datatype

        /* a skipped macroblock indicating block is not followed by a
           block dispatched datatype nor by a buffer, thus break. */
        if(mdd->skipped) {
            break;
        }

        /* read block dispatched datatype */
        DOL_read((void*)port, bdd, sizeof(BlockDispatchedDatatype), p);

        /* check for consistency */
        if( block_number != bdd->block_index ) {
            Error(p,"block number/index consistency check failed.");
        }

        /* store block */
        memcpy( p->local->send_buffer[block_number], bdd->coefficient_data,
                BLOCK_BUFFER_SIZE*sizeof(int));

       /* proceed to next block, finish and send macroblock as soon as
           block_count blocks are processed. */
        block_number++;

    } while(block_number < gdd->block_count);

    control_code = CTRL_NEXT_MB;
    DOL_write((void*)PORT_OUT, &control_code, sizeof(char), p); //(1)
    DOL_write((void*)PORT_OUT, gdd, sizeof(GopDispatchedDatatype), p);//(2)
    DOL_write((void*)PORT_OUT, mdd, sizeof(MbDispatchedDatatype), p);//(3)
    //(1) send control code
    //(2) send gop dispatched datatype
    //(3) send mb dispatched datatype
    
    /* send decododed data only if mb is non-skipped */
    if(!(mdd->skipped)) {
        /* send complete macroblock data */
        DOL_write((void*)PORT_OUT, p->local->send_buffer,
                   12*BLOCK_BUFFER_SIZE*sizeof(int), p);
    }

    return 0;
}

/******************************************
 * additional functions, called by fire() *
 ******************************************/

/**
    Process finishing function.
    Forwards the end_of_sequence message to the following process and detaches
    itself (process will not be fired by the scheduler any more).
*/
void collect_blocks_finish(DOLProcess *p)
{
    char control_code;

     /* we don't need to wait for all end_of_seq 's from all channels, but
       it can be done if we want to make sure all fifos are empty at the
       end of the program. */

    /* in order to finish the programm properly, the following process
       must be informed, and then the process can be detached. */

    control_code = CTRL_END_OF_SEQ;
    DOL_write((void*)PORT_OUT, &control_code, sizeof(char), p);

#ifdef DEBUG
    fprintf(logfile, "   COLLECT BLOCK: detaching dispatch process...\n");
#endif

    DOL_detach(p);

}


