/** @file
    Group of picture collect process.
    Collects the groups of pictures and stores picture by picture in a file.    
*/

#include <stdio.h>

#include "collect_gops.h"

#ifdef DEBUG
    extern FILE *logfile;
    extern FILE *mb_logfile;
#endif

/* file store routines. include after the local state structure is defined! */
#include "store.c"

/**
    Process initialization function.
    Initializes local variables.
*/
void collect_gops_init(DOLProcess *p) 
{
    int i;
    
    /* basic initialization */
    sprintf(p->local->name, "collect gop");
    p->local->gop_count = 0;
    p->local->in_port_count = 0;
    p->local->frame_count = 0;
    
    /* Initiallize Clip table */
    if (!(p->local->Clip=(unsigned char *)malloc(1024)))
        Error(p, "Clip[] malloc failed\n");

    p->local->Clip += 384;

    for (i=-384; i<640; i++)
        p->local->Clip[i] = (i<0) ? 0 : ((i>255) ? 255 : i);
    
}

/**
    Process fire function (called by the scheduler).
    Reads all decoded pictures of a GOP and stores them picture by picture, each
    in a separate file.
*/
int collect_gops_fire(DOLProcess *p) 
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);    
    CREATEPORTVAR(port);
    char control_code;
    int read_buffer_size[3];
    unsigned char *read_buffer[3];
    
    /* build port name from gop_port_count */
    CREATEPORT( port , PORT_IN, 1 , p->local->in_port_count , NUM_OF_GOP_PORTS);
    

    /* loop through all pictures of the current gop (all coming from the same
       process!) */
    for(;;) {
        
        /* read control code */
        DOL_read((void*)port, &control_code, sizeof(char), p);    
        
        if(control_code == CTRL_END_OF_SEQ) {
#ifdef DEBUG
        fprintf(logfile, "GOP COLLECT: end of seq, now finishing...\n");
#endif                
            collect_gops_finish(p);   
            return 0;            
        }
        
        if((control_code == CTRL_END_OF_GOP) ||
           (control_code == CTRL_END_OF_ATT_GOP)) {
#ifdef DEBUG
        fprintf(logfile, "GOP COLLECT: end of gop, now breaking...\n");
#endif                     
            break;
        }
        
        if(control_code != CTRL_NEXT_PIC_OF_GOP) {
            Error(p, "CTRL_NEXT_PIC_OF_GOP expected.");
        }
        
        /* read gop dispatched data type */
        DOL_read((void*)port, gdd, sizeof(GopDispatchedDatatype), p);    
        DOL_read((void*)port, mdd, sizeof(MbDispatchedDatatype), p); // read mb dispatched data type
        
        /* TODO: allocate memory just once */
        
        read_buffer_size[0] = gdd->coded_picture_width * gdd->coded_picture_height;
        read_buffer_size[1] = gdd->chroma_width * gdd->chroma_height;
        read_buffer_size[2] = read_buffer_size[1];

        if (!(read_buffer[0] = (unsigned char *)malloc(read_buffer_size[0])))
            Error(p, "backward_reference_frame[] malloc failed\n");

        if (!(read_buffer[1] = (unsigned char *)malloc(read_buffer_size[1])))
            Error(p, "backward_reference_frame[] malloc failed\n");

        if (!(read_buffer[2] = (unsigned char *)malloc(read_buffer_size[2])))
            Error(p, "backward_reference_frame[] malloc failed\n");

        DOL_read((void*)port, read_buffer[0], read_buffer_size[0] , p);//(1)
        DOL_read((void*)port, read_buffer[1], read_buffer_size[1], p); //(2)
        DOL_read((void*)port, read_buffer[2], read_buffer_size[2], p); //(3)
        //(1) read first color component
        //(2) read second color component        
        //(3) read third color component
        
        /* write current picture to file */
#ifndef INCLUDE_PERFORMANCE
        Write_Frame(p, read_buffer, p->local->frame_count++ );
#endif
        
        free(read_buffer[0]);
        free(read_buffer[1]);
        free(read_buffer[2]);
        
#ifdef DEBUG
        fprintf(logfile, "GOP COLLECT: frame %d read and written to file.\n", 
                        (p->local->frame_count - 1));
#endif         

    }
    
   /* (circularly) increment port number */
    if(p->local->in_port_count < (NUM_OF_GOP_PORTS-1))
        p->local->in_port_count++;
    else
        p->local->in_port_count = 0;
    
    
#ifdef VERBOSE
    fprintf(stdout,"Group of Pictures complete.\n");
#else
    fprintf(stdout,"|");
    fflush(stdout);
#endif
    
    return 0;
}

/******************************************
 * additional functions, called by fire() *
 ******************************************/

/**
    Process finnishing function.
    Frees allocated memory, closes the logfile (if necessary) and detaches
    itself.
*/
void collect_gops_finish(DOLProcess *p)
{
  
        
    /* if all predecessors of the collect_gop process are finished, then
       the program itself is finished. perform the last steps and detach. */

    /* NOTE: it is not necessary to check whether all predecessors are
       finished, because the arrival of one CTRL_END_OF_SEQUENCE implies
       that no more data will arrive through any channel. this is because
       we strictly increment the port numbers. if we would select the port
       numbers with the help of the rtest/wtest functions (no more kahn
       compatible), we would have the check if all predecessors are actually
       finished. */
           
    /* free memory */// No need to free, here is the end of the execution
    //free(p->local->Clip);
    
    /* close logfiles if opened */
#ifdef  DEBUG
#ifndef STDOUT_DEBUG    
    if( logfile != NULL )
        fclose(logfile);
#endif
#endif            
            
#ifdef DEBUG        
            fprintf(logfile, "   COLLECT GOP: detaching dispatch process...\n");
#endif
#ifdef VERBOSE            
            fprintf(stdout, "Ending Program...\n");
#endif

    DOL_detach(p);    

}
