/** @file
    Macroblock collect process.
    Collects the macroblocks, forms the prediction (based on the
    motion vectors) and adds the prediction and the decoded macroblocks.    
*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>

#include "collect_mb.h"

#ifdef DEBUG
    extern FILE *logfile;
#endif

/** prediction forming routines */
#include "predict_mb.c"

/**
    Process initialization function.
    Initializes local variables.
*/
void collect_mb_init(DOLProcess *p) 
{
    int i;

    /* general initialization */
    sprintf(p->local->name, "collect mb");    
    p->local->in_port_count = 0;
    p->local->first_mb = true;
    p->local->pic_end_flag = true;    /* simulate end of picture to ensure that the
                                         first picture is seen as consistent */
    p->local->first_pic_of_gop = true;
    
    /* Initiallize Clip table */
    if (!(p->local->Clip=(unsigned char *)malloc(1024)))
        Error(p, "Clip[] malloc failed\n");

    p->local->Clip += 384;

    for (i=-384; i<640; i++)
        p->local->Clip[i] = (i<0) ? 0 : ((i>255) ? 255 : i);
    
}

/** 
    Process fire function (called by the scheduler).
    Reads data of a single macroblock, forms prediction based on the motion
    vectors, and adds the macroblock data and the prediction. In addition,
    the frames are reordered such as to re-obtain the original decoded order.
    (refer to the official ISO MPEG standard document for details)
*/
int collect_mb_fire(DOLProcess *p) 
{
    
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);    
    CREATEPORTVAR(port);
    char control_code;
    int bx, by;
    int size;   /* picture buffer size */
    int cc;     /* color component */
    int comp;
    
    /* build input port name */
    CREATEPORT( port , PORT_IN, 1 , p->local->in_port_count , NUM_OF_MB_PORTS);

    /* read gop dispatched datatype */
    DOL_read((void*)port, &control_code, sizeof(char), p);    

    if(control_code == CTRL_END_OF_SEQ) {
 
        Send_Frame(p, p->local->backward_reference_frame);  
             
        collect_mb_finish(p);     
        return 0;        
    }

    if(control_code == CTRL_END_OF_GOP) {
        /* send last picture if this gop has no attached B pictures.
           If B pictures were attached, we must omit this step, because the
           last picture is then the first I picture of the next gop (used as a
           reference only) and is displayed by the next gop. */
        Send_Frame(p, p->local->backward_reference_frame);             
    }
    
    
    if((control_code == CTRL_END_OF_GOP) ||
       (control_code == CTRL_END_OF_ATT_GOP)) {
        
        /* mark begin of new gop (needed by frame_reorder) */
        p->local->first_pic_of_gop = true;
        
        /* clear the reference buffers */
        for (cc=0; cc<3; cc++) {
            if (cc == 0)
                size = gdd->coded_picture_width * gdd->coded_picture_height;
            else
                size = gdd->chroma_width * gdd->chroma_height;

            memset(p->local->backward_reference_frame[cc], 0, size);
            memset(p->local->forward_reference_frame[cc], 0, size);
            memset(p->local->auxframe[cc], 0, size);
        }                        
        
        /* pass end_of_gop control code through */
        DOL_write((void*)PORT_OUT, &control_code, sizeof(char), p);
                
        /* (circularly) increment output port number */
        if(p->local->in_port_count < (NUM_OF_MB_PORTS-1))
            p->local->in_port_count++;
        else
            p->local->in_port_count = 0;        
        
        return 0;        
        
    }
    
    if(control_code != CTRL_NEXT_MB) {
        Error(p, "CTRL_NEXT_MB expected. %d instead");
    }
    
    /* read gop dispatched datatype */
    DOL_read((void*)port, gdd, sizeof(GopDispatchedDatatype), p);    
       
    /* (circularly) increment output port number */
    if(p->local->in_port_count < (NUM_OF_MB_PORTS-1))
        p->local->in_port_count++;
    else
        p->local->in_port_count = 0;
    
    /* allocate picture buffer memory. this must be done once immediately
       after the first gop dispatched datatype is read (as we need height
       and width)*/
    if(p->local->first_mb == true) {

        /* loop through all color components (every picture buffer consists of
           three matrixes, one for each component */
        for (cc=0; cc<3; cc++) {

            if (cc == 0)
                size = gdd->coded_picture_width * gdd->coded_picture_height;
            else
                size = gdd->chroma_width * gdd->chroma_height;

            if (!(p->local->backward_reference_frame[cc] = (unsigned char *)malloc(size)))
                Error(p, "backward_reference_frame[] malloc failed\n");

            if (!(p->local->forward_reference_frame[cc] = (unsigned char *)malloc(size)))
                Error(p, "forward_reference_frame[] malloc failed\n");

            if (!(p->local->auxframe[cc] = (unsigned char *)malloc(size)))
                Error(p, "auxframe[] malloc failed\n");

        }
        
        p->local->first_mb = false;      
    }

    /* read mb dispatched datatype */
    DOL_read((void*)port, mdd, sizeof(MbDispatchedDatatype), p);

    
    if(!(mdd->skipped)) {
    
        /* read macroblock data */
        DOL_read((void*)port, p->local->read_buffer, 
                 12*BLOCK_BUFFER_SIZE*sizeof(int), p);
    }
    
    /* derive current macroblock position within picture */
    bx = 16*(mdd->mb_address % gdd->mb_width);
    by = 16*(mdd->mb_address / gdd->mb_width);
    
#ifdef DEBUG
    fprintf(logfile, " MB COLLECT: macroblock collected with coordinates " 
            "(%d, %d; %d; %d).\n", 
            bx, 
            by, 
            mdd->mb_address, 
            (gdd->mb_width * gdd->mb_height));
    fflush(logfile);
#endif    
    
    /* check if the current mb is the first of a new picture. */
    if(mdd->mb_address == 0) {
        
        if(!p->local->pic_end_flag) {
            Error(p,"first mb of picture does not follow last mb of picture!");
        }
        p->local->pic_end_flag = false;

        if (mdd->picture_structure==FRAME_PICTURE && mdd->second_field) {
            /* recover from illegal number of field pictures */
            mdd->second_field = false;
        }
  
        /* prepare buffers for the new picture */
        Update_Picture_Buffers(p);
        
    }
    

    /* form prediction of actual mb and write it into the buffer. intra coded
       macroblocks do not need any prediction. */
    if (!(mdd->macroblock_type & MACROBLOCK_INTRA)) {
        form_predictions(p,
                         bx,
                         by,
                         mdd->macroblock_type,
                         mdd->motion_type, 
                         mdd->PMV, 
                         mdd->motion_vertical_field_select,
                         mdd->dmvector,
                         mdd->stwtype);
    }

    
    /* a skipped mb consists of the prediction alone */
    if(!(mdd->skipped)) {
        /* add all blocks of the current mb to the just calculated prediction */
        for (comp=0; comp < gdd->block_count; comp++) {

            /* Adding prediction and coefficient data */
            Add_Block(p, comp, bx, by, (mdd->macroblock_type & MACROBLOCK_INTRA)==0);
        }
    }
    
    /* check if the current mb is the last of a picture */
    if(mdd->mb_address >= (gdd->mb_width * gdd->mb_height) - 1) {

        /* reorder frames */
        frame_reorder(p, mdd->bitstream_framenum, mdd->sequence_framenum);
        
        /* switch top and bottom field */
        if (mdd->picture_structure != FRAME_PICTURE)
            mdd->second_field = !mdd->second_field;
        
        /* indicate the finished picture (used by the consistency test) */
        p->local->pic_end_flag = true;
    }

    return 0;
}




/******************************************
 * additional functions, called by fire() *
 ******************************************/

/**
    Process finishing function.
    Forwards the end_of_sequence message, frees allocated memory, and detaches
    itself.
*/
void collect_mb_finish(DOLProcess *p)
{

    char control_code;
        
    /* we don't need to wait for all end_of_seq 's from all channels, but
       it can be done if we want to make sure all fifos are empty at the
       end of the program. */   
 
    
    /* in order to finish the programm properly, the following process
       must be informed, and then the process can be detached. */

    control_code = CTRL_END_OF_SEQ;
    DOL_write((void*)PORT_OUT, &control_code, sizeof(char), p);
            
    /* free memory */
    free(p->local->backward_reference_frame[0]);
    free(p->local->backward_reference_frame[1]);
    free(p->local->backward_reference_frame[2]);
    free(p->local->forward_reference_frame[0]);
    free(p->local->forward_reference_frame[1]);
    free(p->local->forward_reference_frame[2]);
    free(p->local->auxframe[0]);    
    free(p->local->auxframe[1]);    
    free(p->local->auxframe[2]);    
        
#ifdef DEBUG        
    fprintf(logfile, "   COLLECT MB: detaching dispatch process...\n");
#endif

    DOL_detach(p);    
}

/**
    Updates the picture buffers.
    This is necessary in order to re-use the buffers whenever possible. This
    way, only 4 picture buffers are necessary.
*/
void Update_Picture_Buffers(DOLProcess *p)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);    
    int cc;              /* color component index */
    unsigned char *tmp;  /* temporary swap pointer */

    for (cc=0; cc<3; cc++) {
        
        /* B pictures do not need to be save for future reference */
        if (mdd->picture_coding_type == B_TYPE) {
          p->local->current_frame[cc] = p->local->auxframe[cc];
        }
        else {
          /* only update at the beginning of the coded frame */
          if (!mdd->second_field) {
            tmp = p->local->forward_reference_frame[cc];

            /* the previously decoded reference frame is stored
               coincident with the location where the backward 
               reference frame is stored (backwards prediction is not
               needed in P pictures) */
            p->local->forward_reference_frame[cc] = p->local->backward_reference_frame[cc];

            /* update pointer for potential future B pictures */
            p->local->backward_reference_frame[cc] = tmp;
          }

          /* can erase over old backward reference frame since it is not used
             in a P picture, and since any subsequent B pictures will use the 
             previously decoded I or P frame as the backward_reference_frame */
          p->local->current_frame[cc] = p->local->backward_reference_frame[cc];
        }

        /* IMPLEMENTATION:
           one-time folding of a line offset into the pointer which stores the
           memory address of the current frame saves offsets and conditional 
           branches throughout the remainder of the picture processing loop */
        if (mdd->picture_structure == BOTTOM_FIELD) {
          p->local->current_frame[cc] += (cc==0) ? 
            gdd->coded_picture_width : gdd->chroma_width;
        }
    }
}

/** 
    Reorders the frames in order to obtain the original decoded frame order.
    A picture that is not needed any more is sent over the channel to the
    following process, and the needed buffers are kept in the appropriate
    buffer (forward- and backward reference buffer, respectively).
*/
void frame_reorder(DOLProcess *p, int Bitstream_Framenum, int Sequence_Framenum)
{
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);    
    
    /* tracking variables to insure proper output in spatial scalability */
    static int Oldref_progressive_frame, Newref_progressive_frame;

    if ((mdd->sequence_framenum != 0) && (p->local->first_pic_of_gop == false)) {
        
        if (mdd->picture_structure==FRAME_PICTURE || mdd->second_field) {
            
            if (mdd->picture_coding_type==B_TYPE) {

                /* send finished frame to the gop collecting process */
          
                Send_Frame(p, p->local->auxframe);

            }
            else {

                Newref_progressive_frame = mdd->progressive_frame;
                mdd->progressive_frame = Oldref_progressive_frame;

                /* send finished frame to the gop collecting process */
                Send_Frame(p, p->local->forward_reference_frame);
                
                Oldref_progressive_frame = mdd->progressive_frame = Newref_progressive_frame;
            }
        }
    }
    else
        Oldref_progressive_frame = mdd->progressive_frame;
    
    if(p->local->first_pic_of_gop == true)
        p->local->first_pic_of_gop = false;

}

/**
    A single frame, coming from the frame reorder function, is sent over the
    channel to the following collecting process.
*/
void Send_Frame(DOLProcess *p, unsigned char *src[3]) 
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    int buffer_size[3];
    char control_code;
    
    buffer_size[0] = gdd->coded_picture_width * gdd->coded_picture_height;
    buffer_size[1] = gdd->chroma_width * gdd->chroma_height;
    buffer_size[2] = buffer_size[1];    
    

    control_code = CTRL_NEXT_PIC_OF_GOP;
    DOL_write((void*)PORT_OUT, &control_code, sizeof(char), p);//(1)
    DOL_write((void*)PORT_OUT, &(p->local->gop_dispatched_datatype), sizeof(GopDispatchedDatatype), p);//(2)
    DOL_write((void*)PORT_OUT, &(p->local->mb_dispatched_datatype), sizeof(MbDispatchedDatatype), p);//(3)
    DOL_write((void*)PORT_OUT, src[0], buffer_size[0], p); //(4)
    DOL_write((void*)PORT_OUT, src[1], buffer_size[1], p); //(5)
    DOL_write((void*)PORT_OUT, src[2], buffer_size[2], p); //(6)
    //(1) send gop dispatched data type
    //(2) send gop dispatched data type
    //(3) send mb dispatched data type
    //(4) send first color component
    //(5) send second color component
    //(6)* send third color component    
}



/** 
 Add block data and prediction.
 Move/add a 8x8-Block from block[comp] to backward_reference_frame.
 Copy reconstructed 8x8 block from block[comp] to current_frame[]
 ISO/IEC 13818-2 section 7.6.8: Adding prediction and coefficient data
 This stage also embodies some of the operations implied by:
    - ISO/IEC 13818-2 section 7.6.7: Combining predictions
    - ISO/IEC 13818-2 section 6.1.3: Macroblock
*/
void Add_Block(DOLProcess *p, int comp, int bx, int by, int addflag)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);    
    int cc,i, j, iincr;
    unsigned char *rfp;
    int *bp;

    /* derive color component index */
    cc = (comp<4) ? 0 : (comp&1)+1; /* color component index */

    if (cc==0) {
        /* luminance */

        if (mdd->picture_structure==FRAME_PICTURE)
            if (mdd->dct_type) {
                /* field DCT coding */
                rfp = p->local->current_frame[0]
                  + gdd->coded_picture_width*(by+((comp&2)>>1)) + bx + ((comp&1)<<3);
                iincr = (gdd->coded_picture_width<<1) - 8;
            }
            else {
                /* frame DCT coding */
                rfp = p->local->current_frame[0]
                  + gdd->coded_picture_width*(by+((comp&2)<<2)) + bx + ((comp&1)<<3);
                iincr = gdd->coded_picture_width - 8;
            }
            else {
                /* field picture */
                rfp = p->local->current_frame[0]
                    + (gdd->coded_picture_width<<1)*(by+((comp&2)<<2)) + bx + ((comp&1)<<3);
                iincr = (gdd->coded_picture_width<<1) - 8;
            }
    }
    else {
        /* chrominance */

        /* scale coordinates */
        if (gdd->chroma_format != CHROMA444)
            bx >>= 1;
    
        if (gdd->chroma_format == CHROMA420)
            by >>= 1;
    
        if (mdd->picture_structure == FRAME_PICTURE) {
      
            if (mdd->dct_type && (gdd->chroma_format!=CHROMA420)) {
                
                /* field DCT coding */
                rfp = p->local->current_frame[cc]
                    + gdd->chroma_width*(by+((comp&2)>>1)) + bx + (comp&8);
                iincr = (gdd->chroma_width<<1) - 8;
            }
            else {
                
                /* frame DCT coding */
                rfp = p->local->current_frame[cc]
                    + gdd->chroma_width*(by+((comp&2)<<2)) + bx + (comp&8);
                iincr = gdd->chroma_width - 8;
            }
        }       
        else {
            
            /* field picture */
            rfp = p->local->current_frame[cc]
                + (gdd->chroma_width<<1)*(by+((comp&2)<<2)) + bx + (comp&8);
            iincr = (gdd->chroma_width<<1) - 8;
        }
    }

    bp = p->local->read_buffer[comp];

    if (addflag) {
    
        for (i=0; i<8; i++) {
            for (j=0; j<8; j++) {
                
                *rfp = p->local->Clip[*bp++ + *rfp];
                rfp++;
            }

            rfp+= iincr;
        }
    }
    else {
        
        for (i=0; i<8; i++) {
            for (j=0; j<8; j++)
                *rfp++ = p->local->Clip[*bp++ + 128];

            rfp+= iincr;
        }
    }

}


