/** @file
    Block dispatch process.
    This process reads the macroblock data and dispatches the
    blocks to the IDCT processes (transform_block). In the case
    of a skipped macroblock, only the motion vectors are dispatched to
    one following process (as a skipped macroblock contains no DCT
    coefficients).    
*/

#include <stdio.h>

#include "dispatch_blocks.h"

#ifdef DEBUG
    extern FILE *logfile;
#endif

/**
    Process initialization function.
    Initializes local variables.
*/    
void dispatch_blocks_init(DOLProcess *p) 
{
    /* basic initializations */
    sprintf(p->local->name, "disp blocks");    
    p->local->out_port_count = 0;
}

/**
    Process fire function, called by the scheduler.
    Reads a macroblock buffer and splits it up into a number of blocks, which
    are sent to the following processes.
*/
int dispatch_blocks_fire(DOLProcess *p) 
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
    char control_code;
    CREATEPORTVAR(port);
    
    /* read control code */
    DOL_read((void*)PORT_IN, &control_code, sizeof(char), p);
    
    if(control_code == CTRL_END_OF_SEQ) {
        dispatch_blocks_finish(p);
        return 0;
    }    
    
    if((control_code == CTRL_END_OF_GOP) || 
       (control_code == CTRL_END_OF_ATT_GOP)){
        /* the end of the current gop is detected. this information
           must be passed through down to the collect_gops process, for it to
           know when to increment the gop port number. */

        /* build output port name */
        CREATEPORT(port, PORT_OUT, 1, p->local->out_port_count,
                   NUM_OF_BLOCK_PORTS);

        /* (circularly) increment output port number */
        if(p->local->out_port_count < (NUM_OF_BLOCK_PORTS-1))
            p->local->out_port_count++;
        else
            p->local->out_port_count = 0;        

        /* send control code */
        DOL_write((void*)port, &control_code, sizeof(char), p);
        
        return 0;
    }
    
    if(control_code != CTRL_NEXT_MB) {
        Error(p, "CTRL_NEXT_MB expected.");
    }
    
    DOL_read((void*)PORT_IN, gdd, sizeof(GopDispatchedDatatype), p); //(1)
    DOL_read((void*)PORT_IN, mdd, sizeof(MbDispatchedDatatype), p); //(2)
    // (1) read gop dispatched datatype
    // (2) read mb dispatched datatype (contains buffer size)
    
    if(!(mdd->skipped)) {
    
        /* calculate integer buffer size. it is always dividable by sizeof(int)
           as it was multiplied in the dispatch_mb process before being sent. 
           care must be taken if the two processes are to be run on different
           hardware with different integer sizes. */
        p->local->read_buffer_size = (mdd->send_buffer_size)/(sizeof(int));

        /* check buffer size */
        if(p->local->read_buffer_size > MAX_MB_BUFFER_SIZE)
            Error(p, "read buffer size exceeded.");

        /* read mb buffer */
        DOL_read((void*)PORT_IN, p->local->read_buffer, mdd->send_buffer_size , p);
   
    }
    
    /* dispatch blocks */
    dispatch(p);
    
    return 0;
}

/******************************************
 * additional functions, called by fire() *
 ******************************************/

/**
    Process finishing function.
    Forwards the end_of_sequence message to the following processes and detaches
    itself (process will not be fired by the scheduler any more).
*/
void dispatch_blocks_finish(DOLProcess *p)
{
    CREATEPORTVAR(port);
    char control_code;
        
    /* in order to finish the programm properly, all following processes
       must be informed, and then the process can be detached. */
    
    control_code = CTRL_END_OF_SEQ;

    for (int i = 0; i < NUM_OF_BLOCK_PORTS; i++) {
        
        /* build port name from gop_port_count */
        CREATEPORT( port , PORT_OUT, 1 , p->local->out_port_count , NUM_OF_BLOCK_PORTS);

        /* send control code */
        DOL_write((void*)port, &control_code, sizeof(char), p);

         /* (circularly) increment port number */
        if(p->local->out_port_count < (NUM_OF_BLOCK_PORTS-1))
            p->local->out_port_count++;
        else
            p->local->out_port_count = 0;            
                        
    }
    
            
#ifdef DEBUG        
            fprintf(logfile, "  BLOCK DISPATCH: detaching dispatch process...\n");
#endif

    DOL_detach(p);    
}

/**
    Dispatches the current macroblock.
    Macroblock data is read from the read_buffer and for each block, a
    BlockDispatchedDatatype is filled and sent to the following process.
*/
void dispatch(DOLProcess *p)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
    BlockDispatchedDatatype *bdd = &(p->local->block_dispatched_datatype);
    int *read_buffer = p->local->read_buffer;    
    CREATEPORTVAR(port);
    char control_code;
    int i, j;
    
    control_code = CTRL_NEXT_BLOCK;
    
    if(mdd->skipped) {
         
        /* build output port name */
        CREATEPORT( port , PORT_OUT, 1 , p->local->out_port_count , NUM_OF_BLOCK_PORTS);

        /* (circularly) increment output port number */
        if(p->local->out_port_count < (NUM_OF_BLOCK_PORTS-1))
            p->local->out_port_count++;
        else
            p->local->out_port_count = 0;        

        DOL_write((void*)port, &control_code, sizeof(char), p); //(1)
        DOL_write((void*)port, gdd, sizeof(GopDispatchedDatatype), p);//(2)
        DOL_write((void*)port, mdd, sizeof(MbDispatchedDatatype), p);//(3)
        //(1) send control code
        //(2) send gop dispatched datatype
        //(3) send mb dispatched datatype
    }
    else {
        
        /* check if buffer size is correct and buffer contains all blocks */
        if(p->local->read_buffer_size != ((gdd->block_count)*BLOCK_BUFFER_SIZE))
            Error(p, "number or sizes of blocks in macroblock incorrect");

        /* loop through all blocks in macroblock */
        for (i = 0; i < gdd->block_count; i++) {

            /* store block_index */
            bdd->block_index = i;

            /* fill coefficient data buffer */
            for (j = 0; j < BLOCK_BUFFER_SIZE; j++) {

                bdd->coefficient_data[j] = *read_buffer++;        

            }

            /* send block */

            /* build output port name */
            CREATEPORT( port , PORT_OUT, 1 , p->local->out_port_count , NUM_OF_BLOCK_PORTS);

            /* (circularly) increment output port number */
            if(p->local->out_port_count < (NUM_OF_BLOCK_PORTS-1))
                p->local->out_port_count++;
            else
                p->local->out_port_count = 0;
            
            DOL_write((void*)port, &control_code, sizeof(char), p);//(1)
            DOL_write((void*)port, gdd, sizeof(GopDispatchedDatatype), p);//(2)
            DOL_write((void*)port, mdd, sizeof(MbDispatchedDatatype), p);//(3)
            DOL_write((void*)port,bdd,sizeof(BlockDispatchedDatatype),p);//(4)
            //(1) send control code
            //(2) send gop dispatched datatype
            //(3) send mb dispatched datatype
            //(4) send block dispatched datatype
        }
    }

    
}

