/** @file
    Group of picture dispatch process.
    This process reads the MPEG-2 bitstream from the specified file,
    demultiplexes the video stream (and discards the rest). The process then
    sends each GOP to a different channel.
*/

#include <stdio.h>
#include <string.h>

#include "dispatch_gops.h"

#ifdef DEBUG
    FILE *logfile;
#endif

/** 
    Process initialization function.
    Opens the input file, reads videosequence header and initializes the
    decoder.
*/
void dispatch_gops_init(DOLProcess *p) {
   
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    int return_value;
    int code;
   
#ifdef DEBUG
#ifndef STDOUT_DEBUG
    if ((logfile=fopen(LOG_FILE_NAME,"w"))==NULL) {
        Error(p, "Unable to open log file.\n");
    }
#else
    logfile = stdout;
#endif
#endif
        
    /* basic initialization */
    sprintf(p->local->name, "gop dispatch");
    p->local->gop_count = 0;
    p->local->gop_port_count = 0;
    p->local->prev_gop_port = -1;   /* first gop has no previous gop */
    p->local->attach_b_pics = false;
    gdd->store_tga_format = true;
    
    /* choose idct */
    gdd->ref_idct = USE_REF_IDCT;   /* IDCT type is chosen at compile-time */
    
    /* make sure we are on a 32bit integer platform. the get_bits() functions
       require storing 32bit into an int */
    if(sizeof(int) < 4)
        Error(p, "Architecture must support 32bit integers.");

#ifdef INPUT_FILE_QUERY
   printf("Enter the filename: ");
   scanf("%s", input_filename);
#else
    char input_filename[50];
    strcpy(input_filename, "test.m2v");
#endif

    /* open input file */
    if ((p->local->infile=fopen(input_filename,"rb"))==NULL) {
        Error(p, "Unable to open input file.\n");
    }
    
    /* set buffer pointers and fill buffer */
    Initialize_Buffer(p);

    /* The sync byte 0x47 marks the begin of a transport stream */
    if(Show_Bits(p, 8)==0x47) {
        Error(p, "This decoder does not parse transport streams.");
    }
    
    /* search the next start code (0x000001yy) */
    next_start_code(p);
    
    /* get the value of the start code (0x01yy) */
    code = Show_Bits(p, 32);

    /* first assume no system stream */
    p->local->System_Stream_Flag = 0;

    /* identify system stream if present */
    switch(code) {
        case SEQUENCE_HEADER_CODE:
            break;
        case PACK_START_CODE:
        case VIDEO_ELEMENTARY_STREAM:
            p->local->System_Stream_Flag = 1;
            break;
        default:
            Error(p, "Unable to recognize stream type.");
            break;
    }
    
#ifdef DEBUG
    if(p->local->System_Stream_Flag)
        fprintf(logfile, "GOP DISPATCH: System Stream detected\n");
    else
        fprintf(logfile, "GOP DISPATCH: Pure Video Stream detected\n");
#endif

    /* set file pointer back to the beginning */
    fseek(p->local->infile, 0l, 0);
    
    /* set buffer pointers and fill buffer */
    Initialize_Buffer(p); 
    
    /* go to next start code */
    next_start_code(p);
 
    /* get videosequence header */
    if( (return_value = get_videosequence_header(p)) != 1)
         Error(p, "Sequence header code expected.");
    
    /* go to next start code */
    next_start_code(p);
 
    /* If an extension_start_code follows, we have 
       an MPEG-2, otherwise an MPEG-1 stream */
    if ((code = Show_Bits(p, 32))==EXTENSION_START_CODE) {
        
#ifdef DEBUG
        fprintf(logfile, "GOP DISPATCH: MPEG-2 bitstream detected\n");
#endif

        gdd->MPEG2_Flag = 1;
        
        if( (return_value = get_videosequence_extension(p)) != 1)
            Error(p, "sequence_extension_id expected.");
    }
    else {
        
#ifdef DEBUG
        fprintf(logfile, "GOP DISPATCH: MPEG-1 bitstream detected\n");
#endif
        gdd->MPEG2_Flag = 0;

    }
     
    /* initialize video sequence */
    initialize_sequence(p);
        
}

/** 
    Process fire function, called by the scheduler.
    Parses the MPEG-2 stream and dispatches a gop, a soon as one is complete.
*/
int dispatch_gops_fire(DOLProcess *p) {
    
    int code;

    /* go to next start code */
    next_start_code(p);
    
    /* get next start code value */
    switch(code = Show_Bits(p, 32)) {
        case EXTENSION_START_CODE:
            /* The only extensions of our interest are the 
               quant_matrix_extension which
               occurs only inside a group of pictures, and the 
               sequence_extension, which is preceeded by a sequence header code
               and is processed right after the sequence header. But we need
               to know if the scalable extension is present for decoding the
               slices */
            Flush_Buffer32(p); 
            get_seq_scal_ext(p);
            break;
        case USER_DATA_START_CODE:
            /* user data can be completely ignored, as it has no influence
               on the decoding process */     
            Flush_Buffer32(p); 
            break;
        case GROUP_START_CODE:
            /* gop start detected. dispatch this gop. */
            dispatch_gop(p);            
            break;
        case PICTURE_START_CODE:
            /* print out error message. it is actually not
               an erroneous bitstream if no gops are present, but this decoder
               can not decode it properly at the moment. for a bitstream
               with gops this case must never happen */
            Error(p, "Picture without GOP detected.");
            break;
        case SEQUENCE_END_CODE:
            /* end of sequence */
            dispatch_gops_finish(p);
            return 0;
            break;
        case SEQUENCE_HEADER_CODE:
            /* the sequence header can be repeated any time. all following
               gops must be delivered with the new sequence header data */
            if(get_videosequence_header(p) != 1)
                Error(p,"Sequence header code expected.");
            if((Show_Bits(p, 32) != EXTENSION_START_CODE) |  
               (get_videosequence_extension(p) != 1))
                Error(p,"sequence_extension after sequence_header expected.");
            break;
        default:
            /* print error for debugging reasons */
            Error(p, "Unexpected start code.");
    }

    return 0;
}

/**
    Process finishing function.
    Forwards the end_of_sequence message to the following processes and
    detaches itself (process will not be fired by the scheduler any more).
*/
void dispatch_gops_finish(DOLProcess *p) 
{
    CREATEPORTVAR(port);
    char control_code;
        
    control_code = CTRL_END_OF_SEQ;

    for (int i = 0; i < NUM_OF_GOP_PORTS; i++) {       
    
        /* build port name from gop_port_count */
        CREATEPORT( port , PORT_OUT, 1 , p->local->gop_port_count , NUM_OF_GOP_PORTS);

        /* send control code */
        DOL_write((void*)port, &control_code, sizeof(char), p);
        
        /* increment port number */
        if(p->local->gop_port_count < (NUM_OF_GOP_PORTS-1))
            p->local->gop_port_count++;
        else
            p->local->gop_port_count = 0;              
        
    }
    
#ifdef DEBUG        
            fprintf(logfile, "GOP DISPATCH: sequence_end_code detected, detaching dispatch process.\n");
#endif
   
    /* close input file */ 
    fclose(p->local->infile);
          
    DOL_detach(p);
}


/** 
    Dispatches the current gop. 
    Sends the header information, followed by all (encoded) pictures.
    Each picture is preceded by an integer value to indicate the size of the
    following picture buffer. After the last picture of a gop, the port number
    is incremented.
*/
void dispatch_gop(DOLProcess *p)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);    
    int b_size; // send buffer size
    unsigned char *send_buffer;
    int current_byte, next_code;
    int pic_number;
    CREATEPORTVAR(port);
    CREATEPORTVAR(p_port); // previous port
    char control_code;

    /* increment the gop counter, reset the picture counter */
    p->local->gop_count++;
    pic_number = 0;
    
    /* point to begin of buffer array */
    send_buffer = p->local->send_buffer;
        
    /* initialize buffer size */
    b_size = 0;

    /* build port name from gop_port_count */
    CREATEPORT(port, PORT_OUT, 1, p->local->gop_port_count, NUM_OF_GOP_PORTS);
    CREATEPORT(p_port, PORT_OUT, 1, p->local->prev_gop_port, NUM_OF_GOP_PORTS);
    
    /* advance buffer as the start code is read out already */
    Flush_Buffer32(p);
    
    /* extract gop header information, ignore time_code, broken_link 
       and user_data */
    Flush_Buffer(p, 25);
    gdd->closed_gop = Get_Bits(p, 1);
    Flush_Buffer(p, 1);

    /* if this is not the first gop of the sequence and this gop is open,
       we have to attach the first consecutive B-pictures following the
       first I-picture of the next gop to the current gop. */
    if(( gdd->closed_gop == 0 ) &&
       (p->local->gop_count > 1)) {
        p->local->attach_b_pics = true;
    } else {
        p->local->attach_b_pics = false;
    }        
    
    next_start_code(p);
    
    while (Show_Bits(p, 32) == USER_DATA_START_CODE) {
        Flush_Buffer32(p);
        next_start_code(p);        
    }
 
#ifdef DEBUG
    if(gdd->closed_gop) {
        fprintf(logfile, "GOP DISPATCH: sending dispatched datatype of " 
                "a CLOSED gop %d...\n", p->local->gop_count);
    } else {
        fprintf(logfile, "GOP DISPATCH: sending dispatched datatype of " 
                "an OPEN gop %d...\n", p->local->gop_count);
    }
#endif           
    
    control_code = CTRL_NEXT_GOP;
    DOL_write((void*)port, &control_code, sizeof(char), p); // (1)
    DOL_write((void*)port, &(p->local->gop_dispatched_datatype), sizeof(GopDispatchedDatatype), p); // (2)
    DOL_write((void*)port, &(p->local->quantizer_matrices), sizeof(QuantizerMatrices), p); // (3)
    // (1) send control code
    // (2) send dispatched data type 
    // (3) send quantizer matrices
    
    /* the next start code must now be a picture start code. */
    if(Show_Bits(p,32) != PICTURE_START_CODE)
        Error(p, "Picture start code expected.");

    /* NOTE: the following loop can be optimized by making use of Get_Bits32()
       which should be faster than Get_Bits(p, 32). However, the start codes 
       are only guaranteed to be byte-aligned, not word-aligned! */
    do {
        
        /* check for buffer overflow */
        if(b_size==MAX_GOP_PIC_BUFFER_SIZE)
            Error(p,"Maximum buffer size exceeded.");
        
        /* write current byte into buffer */
        current_byte = Get_Bits(p, 8);
        send_buffer[b_size++] = (unsigned char)current_byte;
        
        /* get next code */
        next_code = Show_Bits(p, 32);

        /* the sequence header can be repeated at any time */
        if(next_code == SEQUENCE_HEADER_CODE) {
            
            if(get_videosequence_header(p) != 1)
                Error(p, "repeated sequence header expected.");
            
            get_videosequence_extension(p);

            /* skip the additional extensions, if available.
               (only sequence_display_extension and sequence_scalable_extension
               are allowed here, both of which are not of our interest) */
            next_code = Show_Bits(p, 32);            
            while ((next_code == EXTENSION_START_CODE) || 
                   (next_code == USER_DATA_START_CODE)) {
                Flush_Buffer32(p);
                next_start_code(p);                
                next_code = Show_Bits(p, 32);
            }
        }
        else if(next_code == PICTURE_START_CODE) {
            
            /* a new picture is coming. send the old one before continuing */
            pic_number++;

#ifdef DEBUG
            fprintf(logfile, "    GOP DISPATCH: GOP %d picture %d sent to "
                    "port %x pre_port %x\n",
                    p->local->gop_count, pic_number, port, p_port);
#endif
            
            /* attach the B pictures to the previous gop if necessary */
            if((p->local->attach_b_pics) && (NUM_OF_GOP_PORTS > 1)) {
                /* if we have to attach the b pictures, we must:
                    - send the first I picture of the gop the the current AND
                      the previous gop port
                    - send all consecutive B pictures immediately following the
                      first I picture ONLY to the previous gop port
                    - send all other pictures ONLY to the current gop port
                */
                
                if(pic_number == 1) {
                    /* the first picture of a gop must be an I picture */
                    if ((send_buffer[5] & 0x38) != I_TYPE << 3) {
                        Error(p, "First picture of gop must be an I picture");
                    }
                                            

                    control_code = CTRL_ATTACHED_I_PIC; //(1)
                    DOL_write((void*)p_port, &control_code, sizeof(char), p);
                    DOL_write((void*)port, &b_size, sizeof(int), p);//(2)
                    DOL_write((void*)p_port, &b_size, sizeof(int), p);
                    DOL_write((void*)port, send_buffer, b_size, p); //(3)
                    DOL_write((void*)p_port, send_buffer, b_size, p);
                    //(1) send control code
                    //(2) send gop picture buffer size
                    //(3) send gop buffer                                  
                }
                else {
                    if((send_buffer[5] & 0x38) == B_TYPE << 3) {
                        /* this picture belongs to the first consecutive B
                           pictures immediately following the first I
                           picture */
                        control_code = CTRL_ATTACHED_B_PIC; //(1)
                        DOL_write((void*)p_port,&control_code,sizeof(char),p);
                        DOL_write((void*)p_port, &b_size, sizeof(int), p);//(2)
                        DOL_write((void*)p_port, send_buffer, b_size, p);//(3)
                        // (1) send control code
                        // (2) send gop picture buffer size
                        // (3) send gop buffer
                    }
                    else {
                        /* a picture other than B type means that from now on,
                           we must not attach anything more to the 
                           previous gop. */
                        
                        p->local->attach_b_pics = false;

                        DOL_write((void*)port, &b_size, sizeof(int), p);//(1)
                        DOL_write((void*)port, send_buffer, b_size, p); //(2)
                        // (1) send gop picture buffer size
                        // (2) send gop buffer
                    }
                    
                }
                
                
            }
            else {
                /* if we do not have to attach the b pictures, we can simply
                   send all pictures over the current channel. */
                DOL_write((void*)port, &b_size, sizeof(int), p); //(1)
                DOL_write((void*)port, send_buffer, b_size, p);  //(2) 
                // (1) send gop picture buffer size
                // (2) send gop buffer

            }
            
            /* reset buffer */
            send_buffer = p->local->send_buffer;
            b_size = 0;                   
            
        }
        
    } while((next_code!=GROUP_START_CODE) && (next_code!=SEQUENCE_END_CODE));
    
    /* send last picture (is in buffer but not sent yet), 
       followed by the gop end code 'buffer_size = 0'. */
    DOL_write((void*)port, &b_size, sizeof(int), p);
    DOL_write((void*)port, send_buffer, b_size, p);
    b_size = 0;
    DOL_write((void*)port, &b_size, sizeof(int), p);
    
    /* save port number as previous port number (needed for open gops) */
    p->local->prev_gop_port = p->local->gop_port_count;
    
    /* (circularly) increment port number
       NOTE: instead of just incrementing, the wtest api function could
       be used in order to find an empty fifo, but the model of computation
       is then no longer a real Kahn process network (and a sequence number
       must then be added for maintaining the right gop order). */
    if(p->local->gop_port_count < (NUM_OF_GOP_PORTS-1))
        p->local->gop_port_count++;
    else
        p->local->gop_port_count = 0;

#ifdef DEBUG
    fprintf(logfile, "GOP DISPATCH: GOP no.%d dispatched.\n",
            p->local->gop_count);
#endif
#ifdef VERBOSE
    fprintf(stdout, "GOP no.%d dispatched (closed_gop = %d).\n",
            p->local->gop_count, gdd->closed_gop);
#endif    
    
}

/** 
    Reads out the videosequence header.
    Values that are not of our interest must be read out but can be discarded
    immediately.
*/
int get_videosequence_header(DOLProcess *p)
{
    ReadBuffer *bfr = &(p->local->read_buffer);
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    QuantizerMatrices *qm = &(p->local->quantizer_matrices);
    int i;
    int pos;
    int marker;
    int dummy;
    int code;
    bool load_intra_quantizer_matrix;
    bool load_non_intra_quantizer_matrix;
   
    /* get start code value */
    code = Get_Bits32(p);
  
    /* proceed only if begin of a sequence header */
    if(code == SEQUENCE_END_CODE)
        return -1;
    else if(code != SEQUENCE_HEADER_CODE)
        return 0;    
        
    /* read out and store sequence header values. 
       NOTE: the marker bit is a single '1' that prevents the occurence of
       unwanted start codes. It can be discarded after being read out */
    pos = bfr->Bitcnt;
    gdd->horizontal_size = Get_Bits(p, 12);
    gdd->vertical_size = Get_Bits(p, 12);
  
    dummy = Get_Bits(p, 4);                 /* aspect_ratio_information */
    dummy = Get_Bits(p, 4);                 /* frame_rate_code */
    dummy  = Get_Bits(p, 18);               /* bit_rate_value */
    marker = Get_Bits(p, 1);                /* marker bit */
    dummy = Get_Bits(p, 10);                /* vbv_buffer_size */
    dummy = Get_Bits(p, 1);                 /* constrained_parameters_flag */

    if((load_intra_quantizer_matrix = Get_Bits(p, 1))) {
        for (i=0; i<64; i++)
            qm->intra_quantizer_matrix[scan[ZIG_ZAG][i]] = Get_Bits(p, 8);
    }
    else {
        for (i=0; i<64; i++)
            qm->intra_quantizer_matrix[i] = default_intra_quantizer_matrix[i];
    }

    if((load_non_intra_quantizer_matrix = Get_Bits(p, 1))) {
        for (i=0; i<64; i++)
            qm->non_intra_quantizer_matrix[scan[ZIG_ZAG][i]] = Get_Bits(p, 8);
    }
    else {
        for (i=0; i<64; i++)
            qm->non_intra_quantizer_matrix[i] = 16;
    }
  
    for (i=0; i<64; i++) {
        qm->chroma_intra_quantizer_matrix[i] = 
         qm->intra_quantizer_matrix[i];
        qm->chroma_non_intra_quantizer_matrix[i] = 
         qm->non_intra_quantizer_matrix[i];
    }

    return 1;
}

/** 
    Initializes the decoder.
    If we are decoding a MPEG-1 stream, some of the values must be set to
    specific values for having the decoder work properly.
*/
void initialize_sequence(DOLProcess *p)
{

    static int Table_6_20[3] = {6,8,12};
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);

    if (!(gdd->MPEG2_Flag)) {
        gdd->progressive_sequence = 1;
        gdd->chroma_format = CHROMA420;
        gdd->matrix_coefficients = 5;
        /* the rest of the mpeg-1 initialization is done in the parse_picture
           function of the dispatch_mb process. */
    }

    /* calculate number of mb's per picture dimension */
    gdd->mb_width = (gdd->horizontal_size+15)/16;
    
    if(gdd->MPEG2_Flag && !gdd->progressive_sequence)
        gdd->mb_height = 2*((gdd->vertical_size+31)/32);
    else
        gdd->mb_height = (gdd->vertical_size+15)/16;
   
    gdd->coded_picture_width = 16*gdd->mb_width;
    gdd->coded_picture_height = 16*gdd->mb_height;

    gdd->chroma_width = (gdd->chroma_format == CHROMA444) ? 
        gdd->coded_picture_width : gdd->coded_picture_width >> 1;
    
    gdd->chroma_height = (gdd->chroma_format != CHROMA420) ? 
        gdd->coded_picture_height : gdd->coded_picture_height >> 1;
    
    gdd->block_count = Table_6_20[gdd->chroma_format-1];

#ifdef DEBUG
       fprintf(logfile, "GOP DISPATCH: seq. initialized: pic size: " 
                        "%d x %d, %d x %d, ch_format: %d, block_count: %d\n", 
                            gdd->coded_picture_width, 
                            gdd->coded_picture_height, 
                            gdd->chroma_width, 
                            gdd->chroma_height, 
                            gdd->chroma_format, 
                            gdd->block_count);
#endif         
    
}

/** 
    Reads out the videosequence extension header.
    Values that are not of our interest must be read out but can be discarded
    immediately.    
*/
int get_videosequence_extension(DOLProcess *p) 
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);

    int marker, dummy;
    int ext_ID;
    
    int horizontal_size_extension;
    int vertical_size_extension;
    int vbv_buffer_size_extension;
    
    /* MPEG-1 streams do not have videosequence extensions */ 
    if(Show_Bits(p, 32) != EXTENSION_START_CODE)
        return 0;
    
    /* advance buffer as the start code is read out already */
    Flush_Buffer32(p);
    
    /* read out the extension id */
    ext_ID = Get_Bits(p, 4);

    /* If we do not get a sequence_extension_id, the bitstream syntax is 
       incorrect */
    if (ext_ID != SEQUENCE_EXTENSION_ID)
        return 0;

    /* get and store the sequence extension header values */
    dummy = Get_Bits(p, 8);         /* profile_and_level_indication */
    gdd->progressive_sequence = Get_Bits(p, 1);
    gdd->chroma_format = Get_Bits(p, 2);
    horizontal_size_extension = Get_Bits(p, 2);
    vertical_size_extension = Get_Bits(p, 2);
    dummy = Get_Bits(p, 12);        /* bit_rate_extension */
    marker = Get_Bits(p, 1);        /* marker bit */
    vbv_buffer_size_extension = Get_Bits(p, 8);
    dummy = Get_Bits(p, 1);         /* low_delay */
    dummy = Get_Bits(p, 2);         /* frame_rate_extension_n */
    dummy = Get_Bits(p, 5);         /* frame_rate_extension_d */

    gdd->horizontal_size = (horizontal_size_extension<<12) | 
     (gdd->horizontal_size&0x0fff);
    gdd->vertical_size = (vertical_size_extension<<12) | 
     (gdd->vertical_size&0x0fff);

    return 1;
}

/** 
    Reads out the sequence scalable extension header.
    As this decoder does not decode scalable streams at the moment, 
    this header is not allowed here, and an error is produced.
*/
void get_seq_scal_ext(DOLProcess *p)
{
   
    if(Show_Bits(p,4) == SEQUENCE_SCALABLE_EXTENSION_ID) {
        /* at the moment, do not decode scalable streams */
        Error(p, "This decoder does not decode scalable streams.");
    }
    
}
            
/** 
    Input Buffer initialization.
    All the input buffer routines are taken from the reference implementation
    of the MSSG without much changes.
*/    
void Initialize_Buffer(DOLProcess *p)
{
  ReadBuffer *bfr = &(p->local->read_buffer);
    
  /* buffer initially is byte-aligned */   
  bfr->Incnt = 0;
  
  /* set read pointer to the end of the buffer */
  bfr->Rdptr = bfr->Rdbfr + 2048;
  
  /* Set readmax to the end of the buffer */
  bfr->Rdmax = bfr->Rdptr;

  bfr->Bfr = 0;
  
  /* fill buffer with valid data */
  Flush_Buffer(p, 0);
}

/** 
    Read next chunk from file and fill into buffer.
    All the input buffer routines are taken from the reference implementation
    of the MSSG without much changes.
*/
void Fill_Buffer(DOLProcess *p)
{
    ReadBuffer *bfr = &(p->local->read_buffer);
    int Buffer_Level;

    Buffer_Level = fread(bfr->Rdbfr, sizeof(char), 2048, p->local->infile);    
    bfr->Rdptr = bfr->Rdbfr;

    if (p->local->System_Stream_Flag)
        bfr->Rdmax -= 2048;
  
    /* end of the bitstream file */
    if (Buffer_Level < 2048)
    {
        /* just to be safe */
        if (Buffer_Level < 0)
        Buffer_Level = 0;

        /* pad until the next to the next 32-bit word boundary */
        while (Buffer_Level & 3)
            bfr->Rdbfr[Buffer_Level++] = 0;

	/* pad the buffer with sequence end codes */
        while (Buffer_Level < 2048)
        {
            bfr->Rdbfr[Buffer_Level++] = SEQUENCE_END_CODE>>24;
            bfr->Rdbfr[Buffer_Level++] = SEQUENCE_END_CODE>>16;
            bfr->Rdbfr[Buffer_Level++] = SEQUENCE_END_CODE>>8;
            bfr->Rdbfr[Buffer_Level++] = SEQUENCE_END_CODE&0xff;
        }
    }
}


/** 
    Advances the buffer by N bits (discarding the contents).
    All the input buffer routines are taken from the reference implementation
    of the MSSG without much changes.
*/
void Flush_Buffer(DOLProcess *p, int N)
{
  int Incnt;
  ReadBuffer *bfr = &(p->local->read_buffer);

  bfr->Bfr <<= N;

  Incnt = bfr->Incnt -= N;

  if (Incnt <= 24)
  {
    if (p->local->System_Stream_Flag && (bfr->Rdptr >= bfr->Rdmax-4))
    {
      do
      {
        if (bfr->Rdptr >= bfr->Rdmax)
          Next_Packet(p);
        bfr->Bfr |= Get_Byte(p) << (24 - Incnt);
        Incnt += 8;
      }
      while (Incnt <= 24);
    }
    else if (bfr->Rdptr < bfr->Rdbfr+2044)
    {
      do
      {
        bfr->Bfr |= *bfr->Rdptr++ << (24 - Incnt);
        Incnt += 8;
      }
      while (Incnt <= 24);
    }
    else
    {
      do
      {
        if (bfr->Rdptr >= bfr->Rdbfr+2048)
          Fill_Buffer(p);
        bfr->Bfr |= *bfr->Rdptr++ << (24 - Incnt);
        Incnt += 8;
      }
      while (Incnt <= 24);
    }
    bfr->Incnt = Incnt;
  }
}

/** 
    Parses the system layer.
    everything we don't need is ignored (audio, etc...).
    All the input buffer routines are taken from the reference implementation
    of the MSSG without much changes.
*/
void Next_Packet(DOLProcess *p)
{
  unsigned int code;
  int l;
  ReadBuffer *bfr = &(p->local->read_buffer);

  for(;;)
  {
    code = Get_Long(p);

    /* remove system layer byte stuffing */
    while ((code & 0xffffff00) != 0x100)
      code = (code<<8) | Get_Byte(p);

    switch(code)
    {
    case PACK_START_CODE: /* pack header */
      /* skip pack header (system_clock_reference and mux_rate) */
      bfr->Rdptr += 8;
      break;
    case VIDEO_ELEMENTARY_STREAM:   
      code = Get_Word(p);             /* packet_length */
      bfr->Rdmax = bfr->Rdptr + code;

      code = Get_Byte(p);

      if((code>>6)==0x02)
      {
        bfr->Rdptr++;
        code=Get_Byte(p);  /* parse PES_header_data_length */
        bfr->Rdptr+=code;    /* advance pointer by PES_header_data_length */
        return;
      }
      else if(code==0xff)
      {
        /* parse MPEG-1 packet header */
        while((code=Get_Byte(p))== 0xFF);
      }
       
      /* stuffing bytes */
      if(code>=0x40)
      {
        if(code>=0x80)
        {
          Error(p, "Error 1 in packet header\n");
        }
        /* skip STD_buffer_scale */
        bfr->Rdptr++;
        code = Get_Byte(p);
      }

      if(code>=0x30)
      {
        if(code>=0x40)
        {
          Error(p, "Error 2 in packet header\n");
        }
        /* skip presentation and decoding time stamps */
        bfr->Rdptr += 9;
      }
      else if(code>=0x20)
      {
        /* skip presentation time stamps */
        bfr->Rdptr += 4;
      }
      else if(code!=0x0f)
      {
        Error(p, "Error 3 in packet header\n");
      }
      return;
    case ISO_END_CODE: /* end */
      /* simulate a buffer full of sequence end codes */
      l = 0;
      while (l<2048)
      {
        bfr->Rdbfr[l++] = SEQUENCE_END_CODE>>24;
        bfr->Rdbfr[l++] = SEQUENCE_END_CODE>>16;
        bfr->Rdbfr[l++] = SEQUENCE_END_CODE>>8;
        bfr->Rdbfr[l++] = SEQUENCE_END_CODE&0xff;
      }
      bfr->Rdptr = bfr->Rdbfr;
      bfr->Rdmax = bfr->Rdbfr + 2048;
      return;
    default:
      if(code>=SYSTEM_START_CODE)
      {
        /* skip system headers and non-video packets*/
        code = Get_Word(p);
        bfr->Rdptr += code;
      }
      else
      {
        fprintf(stderr,"Unexpected startcode %08x in system layer\n",code);
        exit(1);
      }
      break;
    }
  }
}

/** 
    Advances to the next start code (0x000001).
    All the input buffer routines are taken from the reference implementation
    of the MSSG without much changes.
*/
void next_start_code(DOLProcess *p)
{
    ReadBuffer *bfr = &(p->local->read_buffer);

    /* byte align */
    Flush_Buffer(p, bfr->Incnt&7);
    while (Show_Bits(p,24)!=0x01L)
        Flush_Buffer(p, 8);
}


int Get_Long(DOLProcess *p)
{
  int i;

  i = Get_Word(p);
  return (i<<16) | Get_Word(p);
}

int Get_Word(DOLProcess *p)
{
  int Val;

  Val = Get_Byte(p);
  return (Val<<8) | Get_Byte(p);
}

int Get_Byte(DOLProcess *p)
{
  ReadBuffer *bfr = &(p->local->read_buffer);

  /* fill buffer from file if read pointer is at the end of buffer */
  while(bfr->Rdptr >= bfr->Rdbfr+2048)
  {
    fread(bfr->Rdbfr, sizeof(char), 2048, p->local->infile);
    bfr->Rdptr -= 2048;
    bfr->Rdmax -= 2048;
  }
  
  /* get byte and increment read pointer */
  return *bfr->Rdptr++;
}


/** 
    Returns the next N bits (right adjusted) without advancing.
    All the input buffer routines are taken from the reference implementation
    of the MSSG without much changes.
*/
unsigned int Show_Bits(DOLProcess *p, int N)
{
    ReadBuffer *bfr = &(p->local->read_buffer);

    return bfr->Bfr >> (32-N);
}

void Flush_Buffer32(DOLProcess *p)
{
    int Incnt;
    ReadBuffer *bfr = &(p->local->read_buffer);

    bfr->Bfr = 0;

    Incnt = bfr->Incnt;
    Incnt -= 32;

    if (p->local->System_Stream_Flag && (bfr->Rdptr >= bfr->Rdmax-4))
    {
        while (Incnt <= 24)
        {
        if (bfr->Rdptr >= bfr->Rdmax)
            Next_Packet(p);
        bfr->Bfr |= Get_Byte(p) << (24 - Incnt);
        Incnt += 8;
        }
    }
    else
    {
        while (Incnt <= 24)
        {
        if (bfr->Rdptr >= bfr->Rdbfr+2048)
            Fill_Buffer(p);
        bfr->Bfr |= *bfr->Rdptr++ << (24 - Incnt);
        Incnt += 8;
        }
    }
    bfr->Incnt = Incnt;
}


unsigned int Get_Bits32(DOLProcess *p)
{
    unsigned int l;

    l = Show_Bits(p,32);
    Flush_Buffer32(p);

    return l;
}

/** 
    Returns the next N bits (right adjusted) and advances.
    All the input buffer routines are taken from the reference implementation
    of the MSSG without much changes.
*/
unsigned int Get_Bits(DOLProcess *p, int N)
{
  unsigned int Val;

  Val = Show_Bits(p, N);
  Flush_Buffer(p, N);

  return Val;
}

/** 
    Prints out an error message.
    All eventually opened files that are closed here.
*/
void Error(DOLProcess *p, char *text)
{
    fprintf(stderr, "ERROR in <%s>: %s\n\n", p->local->name, text);
    
    /* there are eventually open files */
    if( strcmp( p->local->name, "gop dispatch" ) == 0 )
        fclose(p->local->infile);

#ifdef DEBUG
#ifndef STDOUT_DEBUG
    /* close logfiles if opened */
    if( logfile != NULL )
        fclose(logfile);
#endif
#endif   
    
    exit(1);
   
}
