/** @file
    Macroblock dispatch process.
    This process reads the pictures of a gop, parses all remaining headers, 
    performes all the variable length decoding (motion vectors and 
    dct coefficients) and dispatches packets of single macroblocks.
    Note that the VLD is necessary in order to find the end of a coded
    macroblock. It is anyways desirable to perform the VLD at this early
    stage in order to improve the pipeline utilization (i.e. the vld can be
    performed already while the previous macroblock is still in the 
    inverse-DCT stage.
*/

#include <stdio.h>
#include <string.h>
#include "dispatch_mb.h"

#ifdef DEBUG
    extern FILE *logfile;
#endif

/* the following included .c files contain additional functionally grouped
   routines. they must be included after the type definition of the
   local state and before the first function definitions. */

/* variable length decoding routines */
#include "vlc_mb.c"
/* motion vector routines */
#include "motion_mb.c"

/**
    Process initialization function.
    Initializes local variables.
*/    
void dispatch_mb_init(DOLProcess *p) 
{
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);

    /* basic initializations */
    sprintf(p->local->name, "dispach mb");
    p->local->gop_count = 0;
    
#ifdef DEBUG    
    p->local->pic_count = 0;
    p->local->slice_count = 0;
    p->local->mb_count = 0;
#endif
    
 
    /* select first output port */
    p->local->out_port_count = 0;

    /* initialize frame numbers */
    mdd->bitstream_framenum = 0;
    mdd->sequence_framenum = 0;
    
    /* initialize field */
    mdd->second_field = false;

}

/** 
    Process fire function, called by the scheduler.
    Reads the GOP and parses each slice of each picture in order to
    send each macroblocks over the next channel.
*/
int dispatch_mb_fire(DOLProcess *p) 
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    QuantizerMatrices *qm = &(p->local->quantizer_matrices);
    BitBuffer *bb = &(p->local->bit_buffer);
    CREATEPORTVAR(port);
    char control_code;

    /* allocate memory once for all pictures of the gop */
    bb->buffer = (unsigned char*)malloc(MAX_GOP_PIC_BUFFER_SIZE);
    
    /* read control code */
    DOL_read((void*)PORT_IN, &control_code, sizeof(char), p);
    
    if(control_code == CTRL_END_OF_SEQ) {
        dispatch_mb_finish(p);
        return 0;
    }
    
    /* if the last received gop was an open gop, we must read out all attached
       pictures before proceeding with the next gop. if this is the first gop, 
       no attached pictures are allowed */
    if(p->local->gop_count > 0) {
        
        if(control_code == CTRL_ATTACHED_I_PIC) {

            /* read out and parse the first attached I picture */
            DOL_read((void*)PORT_IN, &bb->buffer_size, sizeof(int), p);
            DOL_read((void*)PORT_IN, bb->buffer, bb->buffer_size, p);

            parse_picture(p);            

            do {
                /* read control code */
                DOL_read((void*)PORT_IN, &control_code, sizeof(char), p);

                if(control_code != CTRL_ATTACHED_B_PIC)
                    break;

                /* read out and parse the B picture */
                DOL_read((void*)PORT_IN, &bb->buffer_size, sizeof(int), p);
                DOL_read((void*)PORT_IN, bb->buffer, bb->buffer_size, p);

                parse_picture(p);                        

            } while(true);
            
            control_code = CTRL_END_OF_ATT_GOP;            
        
        }
        else if (control_code == CTRL_ATTACHED_B_PIC) {
            Error(p, "CTRL_ATTACHED_B_PIC found without preceding I-picture.");       
        }
        else {
            control_code = CTRL_END_OF_GOP;
        }
        
       /* after having read out all eventually attached pictures, we can 
           mark the (true) end of the (closed or open) gop */

        /* build output port name */
        CREATEPORT( port , PORT_OUT, 1 , p->local->out_port_count , NUM_OF_MB_PORTS);

        /* send control code */
        DOL_write((void*)port, &control_code, sizeof(char), p);                

        /* (circularly) increment output port number. */
        if(p->local->out_port_count < (NUM_OF_MB_PORTS-1))
            p->local->out_port_count++;
        else
            p->local->out_port_count = 0;        
         
    } 
    else {

        if(control_code != CTRL_NEXT_GOP) {
            Error(p, "CTRL_NEXT_GOP expected (first).");
        }               
    }
    

    DOL_read((void*)PORT_IN, gdd, sizeof(GopDispatchedDatatype), p); //(1)
    DOL_read((void*)PORT_IN, qm, sizeof(QuantizerMatrices), p); //(2)
    // (1) read dispatched data type from gop dispatch process
    // (2) read quantizer matrices
    
    /* increment GOP counter as soon as the dispatched data type is read */
    p->local->gop_count++;
    
#ifdef DEBUG    
    p->local->pic_count = 0;
#endif
    
#ifdef DEBUG
    fprintf(logfile, " MBDISPATCH (%s): starting with GOP no.%d\n", p->local->id, p->local->gop_count);
#endif           
    
    /* read out pictures one by one until a gop end code '0' arrives */
    do {
                    
#ifdef DEBUG        
            p->local->pic_count++;
            p->local->slice_count = 0;
#endif
         
            /* read out gop picture buffer size */
            DOL_read((void*)PORT_IN, &bb->buffer_size, sizeof(int), p);

            /* a size of '0' indicates the end of the gop. */
            if (bb->buffer_size == 0) {
                break;
            }
             
            /* read out gop picture buffer */
            DOL_read((void*)PORT_IN, bb->buffer, bb->buffer_size, p);
            
            /* parse headers and send macroblocks to out channels */
            parse_picture(p);
        
#ifdef DEBUG
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
    fprintf(logfile, " MBDISPATCH: picture %d of gop %d parsed (type %d).\n", 
        p->local->pic_count, 
        p->local->gop_count, 
        mdd->picture_coding_type);
#endif
            
    } while(true);


#ifdef DEBUG
    fprintf(logfile, " MBDISPATCH (%s): GOP no.%d processed, " 
                     "finishing fire...\n", 
                     p->local->id, 
                     p->local->gop_count);
#endif        
    
    /* free memory */
    free(bb->buffer);
  
#ifdef DEBUG
    fflush(logfile);
#endif

    return 0;    

}

/**
    Process finishing function.
    Forwards the end_of_sequence message to the following processes and detaches
    itself (process will not be fired by the scheduler any more).
*/
void dispatch_mb_finish(DOLProcess *p)
{
    CREATEPORTVAR(port);
    char control_code;
        
    /* in order to finish the programm properly, all following processes
       must be informed, and then the process can be detached. */

    control_code = CTRL_END_OF_SEQ;    
    
    for (int i = 0; i < NUM_OF_MB_PORTS; i++) {       
    
        /* build port name from gop_port_count */
        CREATEPORT( port , PORT_OUT, 1 , p->local->out_port_count , NUM_OF_MB_PORTS);

        /* send control code */
        DOL_write((void*)port, &control_code, sizeof(char), p);
        
        /* increment port number */
        if(p->local->out_port_count < (NUM_OF_MB_PORTS-1))
            p->local->out_port_count++;
        else
            p->local->out_port_count = 0;              
        
    }
  
            
#ifdef DEBUG        
            fprintf(logfile, " MB DISPATCH: detaching dispatch process...\n");
#endif

    DOL_detach(p);    
}

/******************************************
 * additional functions, called by fire() *
 ******************************************/

/**
   Parses picture stream.
   Note that it is necessary to perform all the variable length decoding in 
   order to dispatch the macroblocks, as there are no more start codes on 
   this level. */
void parse_picture(DOLProcess *p)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    QuantizerMatrices *qm = &(p->local->quantizer_matrices);    
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
    BitBuffer *bb = &(p->local->bit_buffer);    
    int dummy, code, i;
    bool load_intra_quantizer_matrix;
    bool load_non_intra_quantizer_matrix;
    bool load_chroma_intra_quantizer_matrix;
    bool load_chroma_non_intra_quantizer_matrix;
    bool composite_display_flag;
    
    
    /* adding picture start code at the end of the buffer. this is
       necessary to avoid that the vlc routines try to read over the 
       buffer borders. */
    if(bb->buffer_size < MAX_GOP_PIC_BUFFER_SIZE - 4) {
        bb->buffer[bb->buffer_size] = 0x00;
        bb->buffer[bb->buffer_size + 1] = 0x00;
        bb->buffer[bb->buffer_size + 2] = 0x01;
        bb->buffer[bb->buffer_size + 3] = 0x00;
        bb->buffer_size += 4;
    }
    else {
        Error(p, "to big picture buffer (unable to add pic start code");
    }

    /* set buffer pointer to the beginning */
    bb->byte_pos = 0;
    bb->bit_pos = 0;    
    
    /* next start code must be picture start code */
    if(bb_Get_Bits(p,32) != PICTURE_START_CODE)
        Error(p,"picture start code expected");
     
    /* get picture header */
    dummy = bb_Get_Bits(p,10); /* temporal_reference */
    mdd->picture_coding_type = bb_Get_Bits(p,3);
    dummy = bb_Get_Bits(p,16); /* vbv_delay */

     
    if (mdd->picture_coding_type==P_TYPE || mdd->picture_coding_type==B_TYPE) {
        p->local->full_pel_forward_vector = bb_Get_Bits(p,1);
        p->local->forward_f_code = bb_Get_Bits(p,3);
    }
    if (mdd->picture_coding_type==B_TYPE) {
        p->local->full_pel_backward_vector = bb_Get_Bits(p,1);
        p->local->backward_f_code = bb_Get_Bits(p,3);
    }

   while (bb_Get_Bits(p,1)) {
       bb_Advance_Bits(p,8);
#ifdef DEBUG
       fprintf(logfile, " MBDISPATCH: non conform bitstream: " 
                        "extra_information_picture detected");
#endif
    }
    
    dummy = bb_next_start_code(p);
    
    /* in MPEG-2, picture coding extension must follow the picture header */
    if(bb_Show_Bits(p,32) == EXTENSION_START_CODE) {
    
        bb_Advance_Bits(p, 32);
        
        /* if an extension follows, it must be picture coding ext. */
        if(bb_Get_Bits(p,4) != PICTURE_CODING_EXTENSION_ID)
            Error(p,"Picture coding extension identifier expected");

        /* get picture coding extension */

        p->local->f_code[0][0] = bb_Get_Bits(p,4);
        p->local->f_code[0][1] = bb_Get_Bits(p,4);
        p->local->f_code[1][0] = bb_Get_Bits(p,4);
        p->local->f_code[1][1] = bb_Get_Bits(p,4);

        p->local->intra_dc_precision    = bb_Get_Bits(p,2);
        mdd->picture_structure          = bb_Get_Bits(p,2);
        mdd->top_field_first            = bb_Get_Bits(p,1);
        p->local->frame_pred_frame_dct  = bb_Get_Bits(p,1);
        p->local->concealment_motion_vectors = bb_Get_Bits(p,1);
        p->local->q_scale_type               = bb_Get_Bits(p,1);
        p->local->intra_vlc_format           = bb_Get_Bits(p,1);
        p->local->alternate_scan             = bb_Get_Bits(p,1);
        dummy                       = bb_Get_Bits(p,1); /* repeat_first_field */
        dummy                       = bb_Get_Bits(p,1); /* chroma_420_type */
        mdd->progressive_frame          = bb_Get_Bits(p,1);
        composite_display_flag     = bb_Get_Bits(p,1);

        if (composite_display_flag) {
          dummy = bb_Get_Bits(p,1); /* v_axis */
          dummy = bb_Get_Bits(p,3); /* field_sequence */
          dummy = bb_Get_Bits(p,1); /* sub_carrier */
          dummy = bb_Get_Bits(p,7); /* burst_amplitude */
          dummy = bb_Get_Bits(p,8); /* sub_carrier_phase */
        }
            
        dummy = bb_next_start_code(p);
        
    }
    else {
        /* if no extension detected, this is an mpeg-1 stream. in this case
           we must initialize some of the mpeg-2 values for the decoder 
           to work properly */
        
        mdd->progressive_frame = 1;
        mdd->picture_structure = FRAME_PICTURE;
        p->local->frame_pred_frame_dct = 1;
    }

    /* calculate the maximum number of macroblocks */
    p->local->mb_address_max = gdd->mb_width * gdd->mb_height;

    if (mdd->picture_structure != FRAME_PICTURE) {
        
        /* field picture has half as many macroblocks as frame */
        p->local->mb_address_max >>= 1;        
    }    
    
    /* parse extension and user data (if available) to 
       obtain quant matrix extension */
    code = bb_Show_Bits(p,32);

    while(check_for_slice(code) != 1) {

        if(code == EXTENSION_START_CODE) {
            bb_Advance_Bits(p,32);
            if(bb_Show_Bits(p, 4) == QUANT_MATRIX_EXTENSION_ID) {
                bb_Advance_Bits(p,4);

                /* load quantizer matrices, if available */
                
                if((load_intra_quantizer_matrix = bb_Get_Bits(p, 1))) {
                    for (i=0; i<64; i++) {
                        qm->chroma_intra_quantizer_matrix[scan[ZIG_ZAG][i]]
                        = qm->intra_quantizer_matrix[scan[ZIG_ZAG][i]]
                        = bb_Get_Bits(p, 8);
                    }
                }

                if((load_non_intra_quantizer_matrix = bb_Get_Bits(p, 1))) {
                    for (i=0; i<64; i++) {
                        qm->chroma_non_intra_quantizer_matrix[scan[ZIG_ZAG][i]]
                        = qm->non_intra_quantizer_matrix[scan[ZIG_ZAG][i]]
                        = bb_Get_Bits(p, 8);
                    }
                }

                if((load_chroma_intra_quantizer_matrix = bb_Get_Bits(p, 1))) {
                    for (i=0; i<64; i++)
                        qm->chroma_intra_quantizer_matrix[scan[ZIG_ZAG][i]] = 
                        bb_Get_Bits(p, 8);
                }

                if((load_chroma_non_intra_quantizer_matrix=bb_Get_Bits(p, 1))) {
                    for (i=0; i<64; i++)
                        qm->chroma_non_intra_quantizer_matrix[scan[ZIG_ZAG][i]] 
                        = bb_Get_Bits(p, 8);
                }
                         
  
            }
            /* the other extensions are ignored (no information of interest) */
        }
        else {
        
            /* user data is ignored (no information of interest) */
            bb_Advance_Bits(p, 32);
            
        }
               
        /* goto next start code while checking for buffer end */
        if(bb_next_start_code(p) == 0) {
            #ifdef DEBUG
            fprintf(stderr, "WARNING: premature end of picture\n");
            #endif
            return;
        }
        /* prepare next start code (for the while condition) */
        code = bb_Show_Bits(p, 32);
    }

    /* we point now to the start code of the first slice of the picture.
       this is the begin of the picture_data. */
    
    /* loop through all slices. */
    while(bb_next_start_code(p)) {
        
        #ifdef DEBUG        
        p->local->slice_count++;
        p->local->mb_count = 0;
        #endif        
        
        if(bb_Show_Bits(p,32) == PICTURE_START_CODE) {
            /* the previously added picture start code indicates
               the end of the buffer */
            break;            
        }
                
        /* parse slice */
        parse_slice(p);

    }
    
    /* the end of the last slice is also the end of the current picture */
    if (!mdd->second_field) {
        mdd->bitstream_framenum++;
        mdd->sequence_framenum++;
    }
    
}

/**
   Parses a slice.
   All the necessary predictor reseting is performed (see official MPEG-2 
   standard document for details), the slice header is read out, and
   parse_mb() is called for each macroblock.
*/
void parse_slice(DOLProcess *p)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
    unsigned int slice_vertical_position, slice_vertical_position_extension;
    int code;
    
    /* get next start code */
    code = bb_Get_Bits(p,32);

    /* next start code must be a slice start code */
    if(check_for_slice(code) != 1) {
        Error(p,"slice start code expected.");
    }
    
    /* reset the dc coefficient predictors to 0 (as in MSSG's mpeg2dec) */
    p->local->dc_dct_predictor[0] = 0;
    p->local->dc_dct_predictor[1] = 0;
    p->local->dc_dct_predictor[2] = 0;
       
    if((p->local->intra_dc_precision > 3) || (p->local->intra_dc_precision < 0))
        Error(p, "intra_dc_precision not valid");
    
    /* reset all motion vector predictors */
    mdd->PMV[0][0][0]=mdd->PMV[0][0][1]=0;
    mdd->PMV[1][0][0]=mdd->PMV[1][0][1]=0;
    mdd->PMV[0][1][0]=mdd->PMV[0][1][1]=0;
    mdd->PMV[1][1][0]=mdd->PMV[1][1][1]=0;        
        
    /* extract vertical position from start code */
    slice_vertical_position = code & 0x000000FF;

    /* calculate the macroblock row, which is the same for all macroblocks
       of the current slice */
    if((gdd->vertical_size > 2800) && (gdd->MPEG2_Flag)) {
        slice_vertical_position_extension = bb_Get_Bits(p,3);
        p->local->mb_row = (slice_vertical_position_extension << 7) +
                       slice_vertical_position - 1;
#ifdef DEBUG
       fprintf(logfile, " MBDISPATCH: vertical position extension detected.\n");
#endif    
        
    }
    else {
        p->local->mb_row = slice_vertical_position - 1;
    }
    
    /* set previous macroblock address to the reset value */
    p->local->prev_mba = (p->local->mb_row * gdd->mb_width) - 1;
    
    /* get quantizer_scale_code. */
    p->local->quantizer_scale_code = bb_Get_Bits(p, 5);
    
    /* calculate quantizer scale factor from quant_scale_code */
    if (gdd->MPEG2_Flag) {
        if (p->local->q_scale_type)
            p->local->quantizer_scale = 
                    Non_Linear_quantizer_scale[p->local->quantizer_scale_code];
        else
            p->local->quantizer_scale = (p->local->quantizer_scale_code << 1);
    }
    else {
        p->local->quantizer_scale = p->local->quantizer_scale_code;
    }

    /* ignore extra information (reserved for future use) */
    if(bb_Show_Bits(p,1) == 1) {
#ifdef DEBUG
       fprintf(logfile, " MBDISPATCH: extra information detected.\n");
#endif        
        bb_Advance_Bits(p, 9);
        while(bb_Show_Bits(p,1) == 1)
            bb_Advance_Bits(p, 9);
    }
   
    /* remove extra_bit_slice */
    bb_Advance_Bits(p, 1);

    /* we are pointing now to the beginning of the first macroblock */
    p->local->first_mb_of_slice = true;

    /* loop all mb's of the current slice */
    do {
        
#ifdef DEBUG        
        p->local->mb_count++;
#endif        
        
        /* parse macroblock */
        parse_mb(p);
        
        /* from now on, if mb_addr_increment is > 1, we have skipped mb's */
        if(p->local->first_mb_of_slice)
            p->local->first_mb_of_slice = false;
        
        /* check if we are at the end of the buffer */
        if(p->local->bit_buffer.byte_pos > p->local->bit_buffer.buffer_size - 4)
            break;
        
        /* if max. number of macroblocks is reached, the picture is decoded */
        if (mdd->mb_address >= p->local->mb_address_max - 1)
            break;
        
    } while(bb_Show_Bits(p,23) != 0);

    /* 23 '0's indicate a new start code, which can only come from the next
       slice. */
     
}

/**
   Parses a macroblock.
   All values are variable length decoded, down to every DCT coefficient, as 
   it is the only way to find out where the next macroblock starts.
   The variable length decoded values of the macroblock is then sent
   via send_mb().
*/
void parse_mb(DOLProcess *p)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
    int i;
    int mba_save;
    
    /* discard MPEG-1 stuffing */
    while(bb_Show_Bits(p, 11) == 0xF) {
        bb_Advance_Bits(p, 11);
#ifdef DEBUG
       fprintf(logfile, " MBDISPATCH: mpeg-1 stuffing removed.\n");
#endif
    }

    /* calculate address increment (count escapes) */
    p->local->mb_address_increment = 0;
    while(bb_Show_Bits(p, 11) == 0x8) {
        bb_Advance_Bits(p, 11);
        p->local->mb_address_increment += 33;
#ifdef DEBUG
        fprintf(logfile, " MBDISPATCH: macroblock escape detected\n");
#endif    
        
    }

    /* variable length decoding of the address increment value */
    p->local->mb_address_increment += vld_mb_address_inc(p);

    /* calculate the current macroblock address */
    mdd->mb_address = p->local->prev_mba + p->local->mb_address_increment;
    
    /* store this address for subsequent recovery (needed for skipped mb's */
    mba_save = mdd->mb_address;

    /* if address increment is > 1 and we are not at the begin of a slice, 
       we have skipped macroblock(s). in that
       case we must send these here before proceeding. it is important in
       b pictures that this is done with the current motion vector predictors */
    if(!(p->local->first_mb_of_slice)) {
        for(i=1; i < p->local->mb_address_increment; i++) {
        
            /* make sure the correct mb_address is sent for skipped mb's. */
            mdd->mb_address = mba_save - p->local->mb_address_increment + i;               

            /* reset dc coeff. predictors */
            p->local->dc_dct_predictor[0]=0;
            p->local->dc_dct_predictor[1]=0;
            p->local->dc_dct_predictor[2]=0;       
            
            /* derive motion_type */
            if (mdd->picture_structure==FRAME_PICTURE)
                mdd->motion_type = MC_FRAME;
            else {
                mdd->motion_type = MC_FIELD;

                mdd->motion_vertical_field_select[0][0] =
                mdd->motion_vertical_field_select[0][1] = 
                    (mdd->picture_structure==BOTTOM_FIELD);
            }

            mdd->stwtype = (mdd->picture_coding_type==I_TYPE) ? 8 : 0;

            /* IMPLEMENTATION: clear MACROBLOCK_INTRA */
            mdd->macroblock_type &= ~MACROBLOCK_INTRA;
  
            if (mdd->picture_coding_type == P_TYPE) {

                /* in the case of p pictures, some mv predictors have to be reset */
                mdd->PMV[0][0][0]=mdd->PMV[0][0][1]=0;
                mdd->PMV[1][0][0]=mdd->PMV[1][0][1]=0;

                /* a buffer size of '0' tells the following process that a skipped 
                   mb arrives, for which the prediction alone shall be used as 
                   picture data. */
                p->local->send_buffer_size = 0;
                p->local->send_buffer_pos = 0;
               
                /* send mb over the channel */
                send_mb(p);

            }
            else if (mdd->picture_coding_type == B_TYPE) {

                /* for b type pictures, the mv predictors must be left unchanged */

                /* a buffer size of '0' indicates a skipped mb */
                p->local->send_buffer_size = 0;
                p->local->send_buffer_pos = 0;
                
                /* send mb over the channel */
                send_mb(p);

            }
            else {
                /* there should be no skipped macroblock in I pictures, except
                   for scalable sequences */
                Error(p, "no skipped mb's allowed in I pictures");
            }
            
        }   /* end of <for all skipped macroblocks> */
    }
    
    /* recover the correct mb_address */
    mdd->mb_address = mba_save;
    
    /* update previous macroblock address for the next mb */
    p->local->prev_mba = mdd->mb_address;
    
    /* get macroblock modes */
    macroblock_modes(p);

    /* get new quantizer scale code, if available */
    if (mdd->macroblock_type & MACROBLOCK_QUANT) {
        p->local->quantizer_scale_code = bb_Get_Bits(p, 5);
        
        /* calculate quantizer scale factor from quant_scale_code */
        if (gdd->MPEG2_Flag) {
            if (p->local->q_scale_type)
                p->local->quantizer_scale = 
                 Non_Linear_quantizer_scale[p->local->quantizer_scale_code];
            else
                p->local->quantizer_scale = (p->local->quantizer_scale_code << 1);
        }
        else {
            p->local->quantizer_scale = p->local->quantizer_scale_code;
        }
    }            

    /* get motion vectors */
    get_mvs(p);

    /* remove marker bit, if available */
    if ((mdd->macroblock_type & MACROBLOCK_INTRA) 
        && p->local->concealment_motion_vectors) {
        bb_Advance_Bits(p, 1);                
    }

    /* get coded block pattern */
    coded_block_patterns(p);

    fill_send_buffer(p);

    /* send the macroblock */
    send_mb(p);
    
}

/**
    Decodes the macroblock modes.
    See official MPEG-2 standard document for details on how to decode
    the modes.
*/
void macroblock_modes(DOLProcess *p)
{
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
    
    /* variable length decoding of the macroblock type */
    mdd->macroblock_type = vld_mb_type(p);

    /* NOTE: we assume here no spatial scalability (an error is
       produced earlier if spatial scalability is used) */

    /* scalable macroblock is not allowed */
    if (mdd->macroblock_type & SPATIAL_TEMPORAL_WEIGHT_CODE_FLAG)
        Error(p,"unexpected scalable macroblock detected");

    /* get spatial_temporal_weight_class for the non scalable case */
    p->local->stwclass = (mdd->macroblock_type & 
                PERMITTED_SPATIAL_TEMPORAL_WEIGHT_CLASS) ? 4 : 0;

    /* get frame/field motion type */
    if (mdd->macroblock_type & 
        (MACROBLOCK_MOTION_FORWARD|MACROBLOCK_MOTION_BACKWARD)) {
        if (mdd->picture_structure == FRAME_PICTURE) {
            mdd->motion_type = p->local->frame_pred_frame_dct ? 
                               MC_FRAME : bb_Get_Bits(p, 2);
        }
        else {
            mdd->motion_type = bb_Get_Bits(p, 2);
        }
    }
    else if ((mdd->macroblock_type & MACROBLOCK_INTRA) && 
             p->local->concealment_motion_vectors) {
        mdd->motion_type = (mdd->picture_structure==FRAME_PICTURE) ? 
                       MC_FRAME : MC_FIELD;
    }

    /* derive motion_vector_count, mv_format and dmv */
    if (mdd->picture_structure==FRAME_PICTURE) {
        p->local->motion_vector_count = 
                  (mdd->motion_type==MC_FIELD && p->local->stwclass<2) ? 2 : 1;
        p->local->mv_format = 
                  (mdd->motion_type==MC_FRAME) ? MV_FRAME : MV_FIELD;
    }
    else {
        p->local->motion_vector_count = (mdd->motion_type==MC_16X8) ? 2 : 1;
        p->local->mv_format = MV_FIELD;
    }

     /* dual prime */
    p->local->dmv = (mdd->motion_type==MC_DMV);

    /* get dct_type (frame DCT / field DCT) */
    mdd->dct_type = (mdd->picture_structure==FRAME_PICTURE)
     && (!p->local->frame_pred_frame_dct)
     && (mdd->macroblock_type & (MACROBLOCK_PATTERN|MACROBLOCK_INTRA))
     ? bb_Get_Bits(p, 1)
     : 0;
    
    /* stwtype is not required necessarily, but its introduction allows the 
       use of the mpeg2dec motion compensation routines without lots of changes */
    mdd->stwtype = (mdd->macroblock_type & 
                PERMITTED_SPATIAL_TEMPORAL_WEIGHT_CLASS) ? 8 : 0;
   
}

/**
    Decodes the motion vectors.
    See official MPEG-2 standard document for details on how to decode
    the motion vectors.
*/
void get_mvs(DOLProcess *p)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
    int mvscale;
    
    /* calculate mvscale */
    mvscale = ((p->local->mv_format==MV_FIELD) 
                && (mdd->picture_structure==FRAME_PICTURE));

    /* decode forward motion vectors */
    if ((mdd->macroblock_type & MACROBLOCK_MOTION_FORWARD) 
    || ((mdd->macroblock_type & MACROBLOCK_INTRA) 
      && p->local->concealment_motion_vectors)) {
         if (gdd->MPEG2_Flag)
             motion_vectors(p, 0, p->local->f_code[0][0]-1, 
                            p->local->f_code[0][1]-1, mvscale);
         else
             motion_vector(p, mdd->PMV[0][0], mdd->dmvector, 
                           p->local->forward_f_code-1, p->local->forward_f_code-1, 0, 
                           0, p->local->full_pel_forward_vector);
    }

    /* decode backward motion vectors */
    if (mdd->macroblock_type & MACROBLOCK_MOTION_BACKWARD) {
        if (gdd->MPEG2_Flag)
            motion_vectors(p, 1, p->local->f_code[1][0]-1, 
                           p->local->f_code[1][1]-1, mvscale);
        else
            motion_vector(p, mdd->PMV[0][1], mdd->dmvector, 
                          p->local->backward_f_code-1, p->local->backward_f_code-1, 0, 0, 
                          p->local->full_pel_backward_vector);
    }
    
}

/**
    Decodes the coded block patterns.
    The coded block patterns define which blocks are present in the 
    current macroblock. See official MPEG-2 standard document for details on 
    how to decode the modes.  
*/
void coded_block_patterns(DOLProcess *p) 
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
    
    if (mdd->macroblock_type & MACROBLOCK_PATTERN) {
        p->local->coded_block_pattern = vld_coded_block_pattern(p);

        if (gdd->chroma_format == CHROMA422) {

            /* coded_block_pattern_1 */
            p->local->coded_block_pattern = (p->local->coded_block_pattern << 2) 
                                       | bb_Get_Bits(p, 2); 

        }
        else if (gdd->chroma_format == CHROMA444) {

            /* coded_block_pattern_2 */
            p->local->coded_block_pattern = (p->local->coded_block_pattern << 6) 
                                       | bb_Get_Bits(p, 6); 
        }
    }
    else {
        p->local->coded_block_pattern = (mdd->macroblock_type & 
            MACROBLOCK_INTRA) ? (1 << gdd->block_count)-1 : 0;
    }    
}


/**
    Prepares and fills the buffer that is sent later over the channel.
    The buffer is filled by decoding all block coefficients of the current
    macroblock.
*/
void fill_send_buffer(DOLProcess *p)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
    int *send_buffer = p->local->send_buffer;
    int i;

    /* prepare buffer. this is necessary because a premature 'end of block' 
       means that the remaining coefficients shall be zero. */
    memset(send_buffer, 0, sizeof(p->local->send_buffer)); 
    p->local->send_buffer_size = 0;
    p->local->send_buffer_pos = 0;
    
    /* loop through all blocks */
    for (i = 0; i < gdd->block_count; i++) {
        
        /* perform variable length decoding. this is necessary here in order to
           find the end of the macroblock. */
 
        if (p->local->coded_block_pattern & (1<<(gdd->block_count-1-i))) {
            if (mdd->macroblock_type & MACROBLOCK_INTRA) {
                if (gdd->MPEG2_Flag) {
                    vld_MPEG2_Intra_Block(p, i);
                }
                else {
                    vld_MPEG1_Intra_Block(p, i);
                }
            }
            else {
                if (gdd->MPEG2_Flag) {
                    vld_MPEG2_Non_Intra_Block(p, i);
                }
                else {
                    vld_MPEG1_Non_Intra_Block(p, i);
                }
            }
            
        }
        else {
            /* if coded_block_pattern indicates that a block is not coded, 
               stuff buffer with zeroes (i.e. increment the pos.) 
               NOTE: skipped blocks are transmitted! this is not optimal as it
               introduced unneccessary communication. in order to avoid this, 
               a method must be elaborated to indicate the following process
               how many blocks are skipped. the difficult part is however, that
               each skipped macroblock has its motion vectors, based on which
               the prediction is formed. this must be taken into account. */
            
            p->local->send_buffer_pos += 64;            
        }
       
    }    

    
    /* update send buffer size */
    p->local->send_buffer_size = p->local->send_buffer_pos;
        
    /* remove marker bit (mpeg-1 d picture only) */
    if(mdd->picture_coding_type==D_TYPE) {
        bb_Advance_Bits(p, 1);                
    }

    /* reset intra_dc predictors */
    if (!(mdd->macroblock_type & MACROBLOCK_INTRA)) {
        p->local->dc_dct_predictor[0]=0;
        p->local->dc_dct_predictor[1]=0;
        p->local->dc_dct_predictor[2]=0;
    }

    /* reset motion vector predictors if intra mb without conceal. mv */
    if ((mdd->macroblock_type & MACROBLOCK_INTRA) && 
        !(p->local->concealment_motion_vectors)) {
            mdd->PMV[0][0][0]=mdd->PMV[0][0][1]=0;
            mdd->PMV[1][0][0]=mdd->PMV[1][0][1]=0;
            mdd->PMV[0][1][0]=mdd->PMV[0][1][1]=0;
            mdd->PMV[1][1][0]=mdd->PMV[1][1][1]=0;
    }

    /* special "No_MC" macroblock_type case (see 7.6.3.5) */
    if ((mdd->picture_coding_type==P_TYPE) 
      && !(mdd->macroblock_type&(MACROBLOCK_MOTION_FORWARD|MACROBLOCK_INTRA))) {
        /* non-intra mb without forward mv in P picture, reset some mv pred. */
        mdd->PMV[0][0][0]=mdd->PMV[0][0][1]=0;
        mdd->PMV[1][0][0]=mdd->PMV[1][0][1]=0;

        /* derive motion_type (see 6.3.17.1: Macroblock modes) */
        if (mdd->picture_structure==FRAME_PICTURE)
            mdd->motion_type = MC_FRAME;
        else {
            mdd->motion_type = MC_FIELD;
            mdd->motion_vertical_field_select[0][0] = 
                (mdd->picture_structure==BOTTOM_FIELD);
        }
    }

  if (p->local->stwclass==4) {
      /* purely spatially predicted macroblock, resetting mv predictions */
        mdd->PMV[0][0][0]=mdd->PMV[0][0][1]=0;
        mdd->PMV[1][0][0]=mdd->PMV[1][0][1]=0;
        mdd->PMV[0][1][0]=mdd->PMV[0][1][1]=0;
        mdd->PMV[1][1][0]=mdd->PMV[1][1][1]=0;
  }
    
}

/**
    Sends the current macroblock over the channel.
    In the case of a skipped macroblock, no buffer is sent. The dispatched 
    datatypes have to be sent anyways, as they contain the motion vectors,
    which are needed to form the prediction of the skippe macroblock.
*/
void send_mb(DOLProcess *p)
{
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
    CREATEPORTVAR(port);
    char control_code;
    
    /* build output port name */
    CREATEPORT( port , PORT_OUT, 1 , p->local->out_port_count , NUM_OF_MB_PORTS);
    
    /* (circularly) increment output port number */
    if(p->local->out_port_count < (NUM_OF_MB_PORTS-1))
        p->local->out_port_count++;
    else
        p->local->out_port_count = 0;
    
    /* send control code */
    control_code = CTRL_NEXT_MB;
    DOL_write((void*)port, &control_code, sizeof(char), p);         

    if(p->local->send_buffer_size == 0){
        /* send a skipped mb */
        
        /* send gop dispatched datatype */
        DOL_write((void*)port, &(p->local->gop_dispatched_datatype), 
         sizeof(GopDispatchedDatatype), p);

        /* indicate mb as skipped */
        mdd->skipped = true;
        
        /* send mb dispatched datatype (contains buffer size) */
        DOL_write((void*)port, mdd, sizeof(MbDispatchedDatatype), p);        

        /* no buffer is sent.
           NOTE: it is unavoidable to 'send' even a skipped macroblock, as
           the motion vectors are necessary for the collect_mb process in order 
           to form the predicton. The buffer, however, can be omitted, since a
           skipped mb does not have any dct coefficients. */           
        
    }
    else {
        /* send a non-skipped (normal) macroblock */

        /* prepare buffer size (in bytes!) */
        mdd->send_buffer_size = (p->local->send_buffer_size) * sizeof(int);    

        /* send gop dispatched datatype */
        DOL_write((void*)port, &(p->local->gop_dispatched_datatype), 
         sizeof(GopDispatchedDatatype), p);

        /* indicate mb as non-skipped (normal) */
        mdd->skipped = false;
        
        DOL_write((void*)port, mdd, sizeof(MbDispatchedDatatype), p); // (1)
        DOL_write((void*)port, p->local->send_buffer, mdd->send_buffer_size , p); //(2)
        // (1) send mb dispatched datatype (contains buffer size) */
        // (2) send mb buffer
    }

}

/** 
    Returns the next N bits without advancing the buffer pointer.
    @param N Number of bits to show.
 */
unsigned int bb_Show_Bits(DOLProcess *p, int N)
{
    BitBuffer *bb = &(p->local->bit_buffer);
    unsigned char *bfr = bb->buffer;
    int byte_pos = bb->byte_pos;
    int bit_pos = bb->bit_pos;
    int buffer_size = bb->buffer_size;
    int remaining_bytes = buffer_size - byte_pos;
    unsigned int b1, b2, b3, b4;
    
    /* ISO/ANSI C only guarantees a minimum size of 16bit for an int value, but
       we checked in dispatch_gops.c if the architecture has 32bit integers. */
    if(N > 32)
        Error(p,"bb_Show_Bits can be used for a maximum of 32 bits.");
    
    /* this buffer implementation allows more than 24 bits only if the buffer
       is byte aligned (i.e. bit_pos = 0). This is not a problem since the
       only 32 bit values contained in the bitstream are the start codes,
       which are always byte aligned. */
    if((N > 24) && (bit_pos != 0))
        Error(p,"more than 24 bits for non-byte aligned buffer not allowed");
   
    /* make sure we don't try to access beyond the buffer */
    if(bb->byte_pos >= bb->buffer_size)
        Error(p, "byte_pos points outside of buffer");

    if (N > ((remaining_bytes)*8 - bit_pos))
        Error(p, "Buffer exceeded in bb_show_bits");
    
    b1 = (unsigned int)(bfr[byte_pos]);
    b2 = (unsigned int)((remaining_bytes > 1) ? bfr[byte_pos+1] : 0);
    b3 = (unsigned int)((remaining_bytes > 2) ? bfr[byte_pos+2] : 0);
    b4 = (unsigned int)((remaining_bytes > 3) ? bfr[byte_pos+3] : 0);
    
    return (((b1 & (0xFF >> (unsigned int)(bit_pos))) << 24 ) 
        | (b2 << 16) | (b3 << 8) | b4) >> (32-(N+bit_pos));
    
}

/** 
    Advances the buffer pointer by N bits.
    \param N Number of bits to advance.
 */
void bb_Advance_Bits(DOLProcess *p, int N)
{
    BitBuffer *bb = &(p->local->bit_buffer);    
    
    /* check if position still inside buffer */
    if(bb->byte_pos >= bb->buffer_size)
        Error(p, "byte_pos points outside of buffer");    
    
    /* advance position indicators */
    bb->byte_pos += (int)(N/8);
    bb->bit_pos += (int)(N%8);
    
    /* increment byte_pos for every 8 bit_pos */
    while(bb->bit_pos > 7) {
        bb->bit_pos -= 8;
        bb->byte_pos++;
    }
    
}

/** 
    Returns N bits and advances the pointer accordingly.
    \param N Number of bits to return and advance.
*/
unsigned int bb_Get_Bits(DOLProcess *p, int N)
{
    unsigned int return_value;
    
    return_value = bb_Show_Bits(p, N);
    bb_Advance_Bits(p, N);

    return return_value;
}

/**
    Advances in the bit buffer until the next occurence of a start code.
    The start codes are indicated by 24 bits containing the value 0x01.
*/
int bb_next_start_code(DOLProcess *p)
{
    BitBuffer *bb = &(p->local->bit_buffer);
    int ret;

    /* byte align */
    ret = bb_Byte_Align(p);
    
    /* search next start code, return 0 if end of buffer */
    for (;;) {
        if(bb->byte_pos > (bb->buffer_size-4)) {
            /* end of buffer reached */
            return 0;
        }
        if(bb_Show_Bits(p,24) == 0x01L) {
            /* next start code found */
            return 1;
        }
        bb_Advance_Bits(p, 8);
    }

}

/**
    Advances the bit buffer to the begin of the next byte.
*/
int bb_Byte_Align(DOLProcess *p) {
    BitBuffer *bb = &(p->local->bit_buffer);
    if (bb->bit_pos == 0) {
        /* buffer is allready byte aligned */
        return 1;        
    }
    else {
        /* align buffer, skip remaining bits */
        bb->bit_pos = 0;
        bb->byte_pos++;
        return 0;
    }
}

/**
    Checks if a start code belongs to a slice.
    The slice start codes are not defined with a specific value, but instead
    with a range of values, which indicate the vertical position of the slice
    in the picture. This function checks, if the start code lies inside this
    range.
    \param c the start code to check.
*/
int check_for_slice(int c)
{
    int val;
    
    /* extract least significant byte */
    val = c & 0xFF;
    
    /* check for slice start code */
    if ((val >= 0x01) && (val <= 0xAF))
        return 1;
    else
        return 0;
      
    
}

