#ifndef GLOBAL_H
#define GLOBAL_H

/*******************************************************************************
 * The following definitions are compile-time configuration and can be changed *
 ******************************************************************************/

/* define if you want debug information */
//#define DEBUG

/* specify the log file (contains debug information */
#define LOG_FILE_NAME "log.txt"

/* define if you want debug information on screen instead of logfile */
#define STDOUT_DEBUG

/* define if you want to specify input file name after program is started.
   undefine if you want to give the input file name as a command line parameter
   (NOTE: the generated systemC code can't interpret command line parameters */
//#define INPUT_FILE_QUERY

/* define if you want more detailed output (... 'out23.tga saved' ... etc) */
#define VERBOSE

/* define the number of parallel processed gops.
   (used for looping all ports)
   IMPORTANT: this number must be less or equal to the number of defined ports
   in the process network specification.
   If there are more processes defined in than actually used,
   the program can not quit correctly, as processes are started that are never
   detached. */

#define NUM_OF_GOP_PORTS        1
#define NUM_OF_MB_PORTS         1
#define NUM_OF_BLOCK_PORTS      1

/* define the idct type (1: reference, 0: fast) */
#define USE_REF_IDCT            0

/* define maximum size buffers
   IMPORTANT: if this values are changed, the fifo sizes of the process network
   specification must be changed as well */
#define MAX_GOP_PIC_BUFFER_SIZE     307200  // 300 kB
#define MAX_MB_BUFFER_SIZE          1024    // 1024*sizeof(int) (int arrays!)
#define BLOCK_BUFFER_SIZE           64      // block coefficient size is fixed

/* specifiy the output file names */
#define OUT_FILE_NAME_F             "out%df"    // frame output
#define OUT_FILE_NAME_A             "out%da"    // top field output
#define OUT_FILE_NAME_B             "out%db"    // bottom field output

/* define if you want one frame per saved image.
   undefine if you want fields seperately */
#define FRAME_STORE

/********************************************************************
 * The following definitions are internal and should not be changed *
 ********************************************************************/

/* define error value (used in vlc routines */
#define ERROR (-1)

/* define profiles and levels */
#define PROFILE_422   (128+5)
#define MAIN_LEVEL    8

#define OBFRSIZE                    4096    // output buffer size (store.c)

/* define control codes */
#define CTRL_END_OF_SEQ         0
#define CTRL_NEXT_GOP           1
#define CTRL_END_OF_GOP         2
#define CTRL_END_OF_ATT_GOP     3
#define CTRL_NEXT_MB            4
#define CTRL_NEXT_BLOCK         5
#define CTRL_NEXT_PIC_OF_GOP    6
#define CTRL_ATTACHED_I_PIC     7
#define CTRL_ATTACHED_B_PIC     8


/* define video start codes */
#define USER_DATA_START_CODE    0x1B2
#define SEQUENCE_HEADER_CODE    0x1B3
#define EXTENSION_START_CODE    0x1B5
#define SEQUENCE_END_CODE       0x1B7
#define GROUP_START_CODE        0x1B8
#define PICTURE_START_CODE      0x100

/* define system start codes */
#define ISO_END_CODE            0x1B9
#define PACK_START_CODE         0x1BA
#define SYSTEM_START_CODE       0x1BB
#define VIDEO_ELEMENTARY_STREAM 0x1E0

/* define extension start code identifiers */
#define SEQUENCE_EXTENSION_ID               1
#define QUANT_MATRIX_EXTENSION_ID           3
#define SEQUENCE_SCALABLE_EXTENSION_ID      5
#define PICTURE_CODING_EXTENSION_ID         8

/* define scalability modes */
#define SC_NONE 0
#define SC_DP   1
#define SC_SPAT 2
#define SC_SNR  3
#define SC_TEMP 4

/* define scanning variants */
#define ZIG_ZAG     0
#define ALTERNATE   1

/* chroma_format */
#define CHROMA420 1
#define CHROMA422 2
#define CHROMA444 3

/* define picture coding types */
#define I_TYPE 1
#define P_TYPE 2
#define B_TYPE 3
#define D_TYPE 4  /* old type, is not supported any more */

/* define picture structure types */
#define TOP_FIELD     1
#define BOTTOM_FIELD  2
#define FRAME_PICTURE 3

/* define motion_type values*/
#define MC_FIELD 1
#define MC_FRAME 2
#define MC_16X8  2
#define MC_DMV   3

/* define mv_format types*/
#define MV_FIELD 0
#define MV_FRAME 1


/* macroblock type bit masks */
#define MACROBLOCK_INTRA                        1
#define MACROBLOCK_PATTERN                      2
#define MACROBLOCK_MOTION_BACKWARD              4
#define MACROBLOCK_MOTION_FORWARD               8
#define MACROBLOCK_QUANT                        16
#define SPATIAL_TEMPORAL_WEIGHT_CODE_FLAG       32
#define PERMITTED_SPATIAL_TEMPORAL_WEIGHT_CLASS 64


/* data type that represents a bitwise buffer for file access */
typedef struct _read_buffer {
  unsigned char Rdbfr[2048];
  unsigned char *Rdptr;
  unsigned char Inbfr[16];
  unsigned int Bfr;
  unsigned char *Rdmax;
  int Incnt;
  int Bitcnt;
} ReadBuffer;

/* data type that represents a buffer with bitwise access */
typedef struct _bit_buffer {
    int buffer_size;
    int byte_pos;
    int bit_pos;
    unsigned char *buffer;
} BitBuffer;

typedef struct _gop_dispatched_datatype {
    /* sequence header */
    int horizontal_size;
    int vertical_size;
    int mb_width;
    int mb_height;
    int coded_picture_width;
    int coded_picture_height;
    int chroma_width;
    int chroma_height;
    int block_count;
    /* sequence extension */
    bool progressive_sequence;
    int chroma_format;
    /* sequence display extension */
    int matrix_coefficients;
    /* group of pictures header */
    bool closed_gop;
    /* general variables */
    bool ref_idct;
    bool MPEG2_Flag;
    bool store_tga_format;
} GopDispatchedDatatype;

typedef struct _quantizer_matrices {
    /* quant matrix extension */
    unsigned char intra_quantizer_matrix[64];
    unsigned char non_intra_quantizer_matrix[64];
    unsigned char chroma_intra_quantizer_matrix[64];
    unsigned char chroma_non_intra_quantizer_matrix[64];
} QuantizerMatrices;

typedef struct _mb_dispatched_datatype {
    /* picture_header() */
    char picture_coding_type;
    /* picture_coding_extension() header */
    char picture_structure;
    bool top_field_first;
    bool progressive_frame;
    /* macroblock */
    int mb_address;
    int macroblock_type;
    char motion_type;
    bool dct_type;
    int motion_vertical_field_select[2][2];
    int motion_code[2][2][2];
    int motion_residual[2][2][2];
    int dmvector[2];
    int PMV[2][2][2];
    int stwtype;
    /* general mb variables */
    int send_buffer_size;
    int second_field;
    int bitstream_framenum;
    int sequence_framenum;
    bool skipped;
} MbDispatchedDatatype;

typedef struct _block_dispatched_datatype {
    int block_index;
    int coefficient_data[BLOCK_BUFFER_SIZE];
} BlockDispatchedDatatype;

#endif
