/** @file
    Routines for writing the decoded output pictures to files.
    Two formats are currently supported, uncompressed TGA format and
    raw YUV format. The YUV format creates 3 files per image, one
    for the luminance and two for the chrominance components, and does
    not contain any information concerning image size etc. TGA is suited
    for visual evaluation of the output, whereas the YUV format can be used
    for an (automated) comparison with a reference implementation.
*/

/* tables */
int Inverse_Table_6_9[8][4] =
{
  {117504, 138453, 13954, 34903}, /* no sequence_display_extension */
  {117504, 138453, 13954, 34903}, /* ITU-R Rec. 709 (1990) */
  {104597, 132201, 25675, 53279}, /* unspecified */
  {104597, 132201, 25675, 53279}, /* reserved */
  {104448, 132798, 24759, 53109}, /* FCC */
  {104597, 132201, 25675, 53279}, /* ITU-R Rec. 624-4 System B, G */
  {104597, 132201, 25675, 53279}, /* SMPTE 170M */
  {117579, 136230, 16907, 35559}  /* SMPTE 240M (1987) */
};


/* prototypes */
void store_one(DOLProcess *, char *, unsigned char *[], int, int, int);
void store_yuv(DOLProcess *, char *,unsigned char *[], int , int , int);
void store_yuv1(DOLProcess *, char *, unsigned char *, int , int , int , int);
void store_ppm_tga(DOLProcess *, char *, unsigned char *[], int ,int , int , int);
void conv422to444(DOLProcess *, unsigned char *, unsigned char *);
void conv420to422(DOLProcess *, unsigned char *, unsigned char *);
void putbyte (DOLProcess *, int);
void putword(DOLProcess *, int w);

/* store a picture as either one frame or two fields */
void Write_Frame(DOLProcess *p, unsigned char *src[3], int frame)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    char outname[256];

#ifdef FRAME_STORE    
    /* if FRAME_STORE is defined, store progressive, independent of how the 
       sequence was coded */
    sprintf(outname,OUT_FILE_NAME_F,frame);
    store_one(p, outname,src,0,gdd->coded_picture_width, gdd->vertical_size);
#else 
    /* if FRAME_STORE is not defined, store it either progressive or interlaced, 
       depending on how the sequence was coded. */

    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);    
    
    if (gdd->progressive_sequence || mdd->progressive_frame) {
        /* progressive */
        sprintf(outname,OUT_FILE_NAME_F,frame);
        store_one(p, outname,src,0,gdd->coded_picture_width, gdd->vertical_size);
    }
    else {
        /* interlaced */
        sprintf(outname,OUT_FILE_NAME_A,frame);
        store_one(p, outname,src,0,gdd->coded_picture_width<<1,gdd->vertical_size>>1);

        sprintf(outname,OUT_FILE_NAME_B,frame);
        store_one(p, outname,src, gdd->coded_picture_width,gdd->coded_picture_width<<1,gdd->vertical_size>>1);
    }
#endif

}


/* store one frame or one field */
void store_one(DOLProcess *p, char *outname, unsigned char *src[3], int offset, int incr, int height)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);

    /* store one frame in yuv format. this function can be extended, if
       additional formats are to be supported. */

    if(gdd->store_tga_format) {
        store_ppm_tga(p, outname,src,offset,incr,height,1);
    }
    else {
        store_yuv(p, outname,src,offset,incr,height);
    }

}

/* separate headerless files for y, u and v */
void store_yuv(DOLProcess *p, char *outname,unsigned char *src[3], int offset, int incr, int height)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    int hsize;
    char tmpname[256];

    hsize = gdd->horizontal_size;

    sprintf(tmpname,"%s.Y",outname);
    store_yuv1(p, tmpname,src[0],offset,incr,hsize,height);

    if (gdd->chroma_format != CHROMA444) {
        offset >>= 1; 
        incr >>= 1; 
        hsize >>= 1;
    }

    if (gdd->chroma_format == CHROMA420) {
        height >>= 1;
    }

    sprintf(tmpname,"%s.U",outname);
    store_yuv1(p, tmpname,src[1],offset,incr,hsize,height);

    sprintf(tmpname,"%s.V",outname);
    store_yuv1(p, tmpname,src[2],offset,incr,hsize,height);
}

/* auxiliary routine */
void store_yuv1(DOLProcess *p, char *name, unsigned char *src, int offset, int incr, int width, int height)
{
    int i, j;
    unsigned char *pt;

    if ((p->local->outfile = fopen(name,"wb"))==NULL) {
        Error(p, "Error creating out file.");
    }

    /* point to the beginning of the output buffer */
    p->local->optr = p->local->obfr;

    for (i=0; i<height; i++) {
        pt = src + offset + incr*i;
        for (j=0; j<width; j++)
            putbyte(p, *pt++);
    }

    if (p->local->optr != p->local->obfr)
        fwrite(p->local->obfr, sizeof(unsigned char), p->local->optr - p->local->obfr, p->local->outfile);

  fclose(p->local->outfile);
}

/* store as PPM (PBMPLUS) or uncompressed Truevision TGA ('Targa') file */
void store_ppm_tga(DOLProcess *p, char *outname, unsigned char *src[3], int offset,int incr, int height, int tgaflag)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    
    int i, j;
    int y, u, v, r, g, b;
    int crv, cbu, cgu, cgv;
    unsigned char *py, *pu, *pv;
    static unsigned char tga24[14] = {0,0,2,0,0,0,0, 0,0,0,0,0,24,32};
    char header[256];
    static unsigned char *u422, *v422, *u444, *v444;

    if (gdd->chroma_format==CHROMA444) {
        u444 = src[1];
        v444 = src[2];
    }
    else {
        if (!u444) {
            if (gdd->chroma_format==CHROMA420) {
                if (!(u422 = (unsigned char *)malloc((gdd->coded_picture_width>>1)
                                             *gdd->coded_picture_height)))
                    Error(p, "    STORE: malloc failed");
                
                if (!(v422 = (unsigned char *)malloc((gdd->coded_picture_width>>1)
                                             *gdd->coded_picture_height)))
                    Error(p, "    STORE: malloc failed");
            }

      
            if (!(u444 = (unsigned char *)malloc(gdd->coded_picture_width
                                           *gdd->coded_picture_height)))
                Error(p, "    STORE: malloc failed");

            if (!(v444 = (unsigned char *)malloc(gdd->coded_picture_width
                                           *gdd->coded_picture_height)))
                Error(p, "    STORE: malloc failed");
        }
    
        if (gdd->chroma_format==CHROMA420) {
            conv420to422(p, src[1],u422);
            conv420to422(p, src[2],v422);
            conv422to444(p, u422,u444);
            conv422to444(p, v422,v444);
        }
        else {
            conv422to444(p, src[1],u444);
            conv422to444(p, src[2],v444);
        }
    }

    strcat(outname,tgaflag ? ".tga" : ".ppm");

#ifdef VERBOSE
    fprintf(stdout," %s saved.\n",outname);
#else
    fprintf(stdout,".");
    fflush(stdout);
#endif

    if ((p->local->outfile = fopen(outname,"wb"))==NULL) {
        Error(p, "Error creating out file.");
    }

    p->local->optr = p->local->obfr;

    if (tgaflag) {
        /* TGA header */
        for (i=0; i<12; i++)
            putbyte(p, tga24[i]);

        putword(p, gdd->horizontal_size); 
        putword(p, height);
        putbyte(p, tga24[12]); 
        putbyte(p, tga24[13]);
    }
    else {
        /* PPM header */
        sprintf(header,"P6\n%d %d\n255\n", gdd->horizontal_size, height);

        for (i=0; header[i]!=0; i++)
            putbyte(p, header[i]);
    }

    /* matrix coefficients */
    crv = Inverse_Table_6_9[gdd->matrix_coefficients][0];
    cbu = Inverse_Table_6_9[gdd->matrix_coefficients][1];
    cgu = Inverse_Table_6_9[gdd->matrix_coefficients][2];
    cgv = Inverse_Table_6_9[gdd->matrix_coefficients][3];

    for (i=0; i<height; i++) {
        py = src[0] + offset + incr*i;
        pu = u444 + offset + incr*i;
        pv = v444 + offset + incr*i;

        for (j=0; j<gdd->horizontal_size; j++) {
            u = *pu++ - 128;
            v = *pv++ - 128;
            y = 76309 * (*py++ - 16); /* (255/219)*65536 */
            r = p->local->Clip[(y + crv*v + 32768)>>16];
            g = p->local->Clip[(y - cgu*u - cgv*v + 32768)>>16];
            b = p->local->Clip[(y + cbu*u + 32786)>>16];

            if (tgaflag) {
                putbyte(p, b); 
                putbyte(p, g); 
                putbyte(p, r);
            }
            else
            {
                putbyte(p, r); 
                putbyte(p, g); 
                putbyte(p, b);
            }
        }
    }     

    if (p->local->optr != p->local->obfr)
        fwrite(p->local->obfr, sizeof(unsigned char), p->local->optr - p->local->obfr, p->local->outfile);

    fclose(p->local->outfile);

}


/* horizontal 1:2 interpolation filter */
void conv422to444(DOLProcess *p, unsigned char *src, unsigned char *dst)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);

    int i, i2, w, j, im3, im2, im1, ip1, ip2, ip3;
  

    w = gdd->coded_picture_width>>1;

    if (gdd->MPEG2_Flag) {
        for (j=0; j<gdd->coded_picture_height; j++) {
            for (i=0; i<w; i++) {
                i2 = i<<1;
                im2 = (i<2) ? 0 : i-2;
                im1 = (i<1) ? 0 : i-1;
                ip1 = (i<w-1) ? i+1 : w-1;
                ip2 = (i<w-2) ? i+2 : w-1;
                ip3 = (i<w-3) ? i+3 : w-1;

                /* FIR filter coefficients (*256): 21 0 -52 0 159 256 159 0 -52 0 21 */
                /* even samples (0 0 256 0 0) */
                dst[i2] = src[i];

                /* odd samples (21 -52 159 159 -52 21) */
                dst[i2+1] = p->local->Clip[(int)(21*(src[im2]+src[ip3])
                                           -52*(src[im1]+src[ip2]) 
                                           +159*(src[i]+src[ip1])+128)>>8];
            }
            src+= w;
            dst+= gdd->coded_picture_width;
        }
    }
    else {
        for (j=0; j<gdd->coded_picture_height; j++) {
            for (i=0; i<w; i++) {

                i2 = i<<1;
                im3 = (i<3) ? 0 : i-3;
                im2 = (i<2) ? 0 : i-2;
                im1 = (i<1) ? 0 : i-1;
                ip1 = (i<w-1) ? i+1 : w-1;
                ip2 = (i<w-2) ? i+2 : w-1;
                ip3 = (i<w-3) ? i+3 : w-1;

                /* FIR filter coefficients (*256): 5 -21 70 228 -37 11 */
                dst[i2] =   p->local->Clip[(int)(  5*src[im3]
                         -21*src[im2]
                         +70*src[im1]
                        +228*src[i]
                         -37*src[ip1]
                         +11*src[ip2]+128)>>8];

                dst[i2+1] = p->local->Clip[(int)(  5*src[ip3]
                         -21*src[ip2]
                         +70*src[ip1]
                        +228*src[i]
                         -37*src[im1]
                         +11*src[im2]+128)>>8];
            }
            src+= w;
            dst+= gdd->coded_picture_width;
        }
    }
}

/* vertical 1:2 interpolation filter */
void conv420to422(DOLProcess *p, unsigned char *src, unsigned char *dst)
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);    
    int w, h, i, j, j2;
    int jm6, jm5, jm4, jm3, jm2, jm1, jp1, jp2, jp3, jp4, jp5, jp6, jp7;

    w = gdd->coded_picture_width>>1;
    h = gdd->coded_picture_height>>1;

    if (mdd->progressive_frame) {
        /* intra frame */
        for (i=0; i<w; i++) {
            for (j=0; j<h; j++) {
                j2 = j<<1;
                jm3 = (j<3) ? 0 : j-3;
                jm2 = (j<2) ? 0 : j-2;
                jm1 = (j<1) ? 0 : j-1;
                jp1 = (j<h-1) ? j+1 : h-1;
                jp2 = (j<h-2) ? j+2 : h-1;
                jp3 = (j<h-3) ? j+3 : h-1;

                /* FIR filter coefficients (*256): 5 -21 70 228 -37 11 */
                /* New FIR filter coefficients (*256): 3 -16 67 227 -32 7 */
                dst[w*j2] =     p->local->Clip[(int)(  3*src[w*jm3]
                             -16*src[w*jm2]
                             +67*src[w*jm1]
                            +227*src[w*j]
                             -32*src[w*jp1]
                             +7*src[w*jp2]+128)>>8];

                dst[w*(j2+1)] = p->local->Clip[(int)(  3*src[w*jp3]
                             -16*src[w*jp2]
                             +67*src[w*jp1]
                            +227*src[w*j]
                             -32*src[w*jm1]
                             +7*src[w*jm2]+128)>>8];
            }
            src++;
            dst++;
        }
    }
    else {
        /* intra field */
        for (i=0; i<w; i++) {
            for (j=0; j<h; j+=2) {
                j2 = j<<1;

                /* top field */
                jm6 = (j<6) ? 0 : j-6;
                jm4 = (j<4) ? 0 : j-4;
                jm2 = (j<2) ? 0 : j-2;
                jp2 = (j<h-2) ? j+2 : h-2;
                jp4 = (j<h-4) ? j+4 : h-2;
                jp6 = (j<h-6) ? j+6 : h-2;
                
        /* Polyphase FIR filter coefficients (*256): 2 -10 35 242 -18 5 */
        /* New polyphase FIR filter coefficients (*256): 1 -7 30 248 -21 5 */
                dst[w*j2] = p->local->Clip[(int)(  1*src[w*jm6]
                         -7*src[w*jm4]
                         +30*src[w*jm2]
                        +248*src[w*j]
                         -21*src[w*jp2]
                          +5*src[w*jp4]+128)>>8];

        /* Polyphase FIR filter coefficients (*256): 11 -38 192 113 -30 8 */
        /* New polyphase FIR filter coefficients (*256):7 -35 194 110 -24 4 */
                dst[w*(j2+2)] = p->local->Clip[(int)( 7*src[w*jm4]
                             -35*src[w*jm2]
                            +194*src[w*j]
                            +110*src[w*jp2]
                             -24*src[w*jp4]
                              +4*src[w*jp6]+128)>>8];

                /* bottom field */
                jm5 = (j<5) ? 1 : j-5;
                jm3 = (j<3) ? 1 : j-3;
                jm1 = (j<1) ? 1 : j-1;
                jp1 = (j<h-1) ? j+1 : h-1;
                jp3 = (j<h-3) ? j+3 : h-1;
                jp5 = (j<h-5) ? j+5 : h-1;
                jp7 = (j<h-7) ? j+7 : h-1;

        /* Polyphase FIR filter coefficients (*256): 11 -38 192 113 -30 8 */
        /* New polyphase FIR filter coefficients (*256):7 -35 194 110 -24 4 */
                dst[w*(j2+1)] = p->local->Clip[(int)( 7*src[w*jp5]
                             -35*src[w*jp3]
                            +194*src[w*jp1]
                            +110*src[w*jm1]
                             -24*src[w*jm3]
                              +4*src[w*jm5]+128)>>8];

                dst[w*(j2+3)] = p->local->Clip[(int)(  1*src[w*jp7]
                             -7*src[w*jp5]
                             +30*src[w*jp3]
                            +248*src[w*jp1]
                             -21*src[w*jm1]
                              +5*src[w*jm3]+128)>>8];
            } 
            src++;
            dst++;
        }
    }
}


void putword(DOLProcess *p, int w)
{
    putbyte(p, w); 
    putbyte(p, w>>8);
}

void putbyte(DOLProcess *p, int c)
{
    *(p->local->optr++) = c;

    if (p->local->optr == p->local->obfr + OBFRSIZE) {
        fwrite(p->local->obfr, sizeof(char), OBFRSIZE, p->local->outfile);
        p->local->optr = p->local->obfr;
    }
}
