/** @file
    Block transform process.
    Transforms a block of DCT coefficients into luminance/chrominance values.
    The process reads blocks of coefficients, transforms them into values and
    forwards the result to the following process.    
*/

#include <stdio.h>

#include "transform_block.h"

#ifdef DEBUG
    extern FILE *logfile;
#endif

/* reference idct routines */
#include "idctref_block.c"
/* fast idct (Chen-Wang) routines */
#include "idct_block.c"

void transform_block_init(DOLProcess *p) 
{
    sprintf(p->local->name, "transf block");    
    
    /* initialize both transform routines */
    Initialize_Reference_IDCT();
    Initialize_Fast_IDCT();
}

int transform_block_fire(DOLProcess *p) 
{
    GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
    BlockDispatchedDatatype *bdd = &(p->local->block_dispatched_datatype);    
    char control_code;
    
    /* read control code */
    DOL_read((void*)PORT_IN, &control_code, sizeof(char), p);
    
    if(control_code == CTRL_END_OF_SEQ) {
        transform_block_finish(p);
        return 0;        
    }
    
    if((control_code == CTRL_END_OF_GOP) || 
       (control_code == CTRL_END_OF_ATT_GOP)){

        /* pass the end of gop flag through */
        DOL_write((void*)PORT_OUT, &control_code, sizeof(char), p);   
        return 0;        
    }
    
    if(control_code != CTRL_NEXT_BLOCK) {
        Error(p, "CTRL_NEXT_BLOCK expected.");
    }
    
    /* read gop dispatched datatype */
    DOL_read((void*)PORT_IN, gdd, sizeof(GopDispatchedDatatype), p);
    DOL_read((void*)PORT_IN, mdd, sizeof(MbDispatchedDatatype), p); // read mb dispatched datatype
    
    if(!(mdd->skipped)) {
        
        /* read block dispatched datatype */
        DOL_read((void*)PORT_IN, bdd, sizeof(BlockDispatchedDatatype), p);

        /* perform inverse discrete cosine transformation */
        if (gdd->ref_idct)
            Reference_IDCT(bdd->coefficient_data);
        else
            Fast_IDCT(bdd->coefficient_data);   
        
    }

    DOL_write((void*)PORT_OUT, &control_code, sizeof(char), p); //(1)
    DOL_write((void*)PORT_OUT, gdd, sizeof(GopDispatchedDatatype), p); //(2)
    DOL_write((void*)PORT_OUT, mdd, sizeof(MbDispatchedDatatype), p); //(3)
    //(1) send control code (is still CTRL_NEXT_BLOCK)
    //(2) send gop dispatched datatype
    //(3) send mb dispatched datatype
    
    if(!(mdd->skipped)) {
        
        /* send block dispatched datatype */
        DOL_write((void*)PORT_OUT, bdd, sizeof(BlockDispatchedDatatype), p);
    
    }

    return 0;
}


/******************************************
 * additional functions, called by fire() *
 ******************************************/

void transform_block_finish(DOLProcess *p)
{
    char control_code;
    
    /* in order to finish the programm properly, the following process
       must be informed, and then this process can be detached. */
    
    control_code = CTRL_END_OF_SEQ;
        
    DOL_write((void*)PORT_OUT, &control_code, sizeof(char), p);
                
#ifdef DEBUG        
    fprintf(logfile, "   TRANSFORM BLOCK: detaching dispatch process...\n");
#endif

    DOL_detach(p);    
}
