/** @file
    Routines for the variable length decoding.
*/

#include "vlc_mb.h"


int vld_mb_type(DOLProcess *p)
{
    int macroblock_type = 0;
    MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
  
    /* we do not consider scalable bitstreams for now */
    
    switch (mdd->picture_coding_type) {
        case I_TYPE:
          macroblock_type = Get_I_macroblock_type(p);
          break;
        case P_TYPE:
          macroblock_type = Get_P_macroblock_type(p);
          break;
        case B_TYPE:
          macroblock_type = Get_B_macroblock_type(p);
          break;
        case D_TYPE:
          macroblock_type = Get_D_macroblock_type(p);
          break;
        default:
          Error(p, "Get_macroblock_type(): unrecognized picture coding type\n");
          break;
    }

    return macroblock_type;
}

static int Get_I_macroblock_type(DOLProcess *p)
{
    if (bb_Get_Bits(p, 1))
        return 1;

    if (!bb_Get_Bits(p, 1))
      Error(p, "Invalid I macroblock_type code");

    return 17;
}

static int Get_P_macroblock_type(DOLProcess *p)
{
    int code;

    if ((code = bb_Show_Bits(p, 6)) >= 8) {
        code >>= 3;
        bb_Advance_Bits(p, PMBtab0[code].len);

        return PMBtab0[code].val;
    }

    if (code==0)
      Error(p, "Invalid P macroblock_type code");

    bb_Advance_Bits(p, PMBtab1[code].len);

    return PMBtab1[code].val;
}

static int Get_B_macroblock_type(DOLProcess *p)
{
    int code;

    if ((code = bb_Show_Bits(p, 6)) >= 8) {
        code >>= 2;
        bb_Advance_Bits(p, BMBtab0[code].len);
        
        return BMBtab0[code].val;
    }

    if (code==0)
        Error(p, "Invalid B macroblock_type code");

    bb_Advance_Bits(p, BMBtab1[code].len);

    return BMBtab1[code].val;
}

static int Get_D_macroblock_type(DOLProcess *p)
{
    if (!bb_Get_Bits(p, 1))
        Error(p, "Invalid D macroblock_type code");
 
    return 1;
}

/* get variable length encoded motion code (table B-10) */
int vld_motion_code(DOLProcess *p)
{
    int code;

    if (bb_Get_Bits(p, 1))
        return 0;
 
    if ((code = bb_Show_Bits(p, 9)) >= 64 ) {

        code >>= 6;
        bb_Advance_Bits(p, MVtab0[code].len);

        /* return positive or negative value, depending on the last bit */
        return bb_Get_Bits(p, 1) ? -MVtab0[code].val : MVtab0[code].val;
    }

    if (code >= 24) {
        
        code >>= 3;
        bb_Advance_Bits(p, MVtab1[code].len);

        /* return positive or negative value, depending on the last bit */
        return bb_Get_Bits(p, 1) ? -MVtab1[code].val : MVtab1[code].val;
    }

    if ((code -= 12) < 0)
        Error(p, "Invalid motion_vector code");
    
    bb_Advance_Bits(p, MVtab2[code].len);

    /* return positive or negative value, depending on the last bit */
    return bb_Get_Bits(p, 1) ? -MVtab2[code].val : MVtab2[code].val;
}

/* get differential motion vector (for dual prime prediction) */
int vld_dmvector(DOLProcess *p)
{
  if (bb_Get_Bits(p, 1)) {

    return bb_Get_Bits(p, 1) ? -1 : 1;
  
  }
  else {

    return 0;
    
  }
}

int vld_coded_block_pattern(DOLProcess *p)
{
    int code;

    if ((code = bb_Show_Bits(p, 9)) >= 128) {
        code >>= 4;
        bb_Advance_Bits(p, CBPtab0[code].len);

        return CBPtab0[code].val;
    }

    if (code >= 8) {
        code >>= 1;
        bb_Advance_Bits(p, CBPtab1[code].len);

        return CBPtab1[code].val;
    }

    if (code<1) {
        Error(p, "Invalid coded_block_pattern code\n");
    }

    bb_Advance_Bits(p, CBPtab2[code].len);

    return CBPtab2[code].val;
}

int vld_mb_address_inc(DOLProcess *p)
{
    int code;
    
    code = bb_Show_Bits(p,11);
    
    /* check for first bit '1', which means address increment of 1 */
    
    if (code >= 1024) {
        bb_Advance_Bits(p,1);
        return 1;
    }

    /* codes 00010 ... 011xx */
    
    if (code >= 128) {
    
        /* remove leading zeros */
        code >>= 6;
        
        bb_Advance_Bits(p, MBAtab1[code].len);

        return MBAtab1[code].val;
    }
  
    /* codes 00000011000 ... 0000111xxxx */
    
    /* remove common base */
    code-= 24; 
    
    if(code < 0)
        Error(p, "wrong mb address inc code");
    
    bb_Advance_Bits(p, MBAtab2[code].len);

    return MBAtab2[code].val;
}

int vld_Luma_DC_dct_diff(DOLProcess *p)
{
  int code, size, dct_diff;

  /* decode length */
  code = bb_Show_Bits(p, 5);

  if (code<31) {
    size = DClumtab0[code].val;
    bb_Advance_Bits(p, DClumtab0[code].len);
  }
  else {
    code = bb_Show_Bits(p, 9) - 0x1f0;
    size = DClumtab1[code].val;
    bb_Advance_Bits(p, DClumtab1[code].len);
  }

  if (size==0)
    dct_diff = 0;
  else {
    dct_diff = bb_Get_Bits(p, size);

    if ((dct_diff & (1<<(size-1)))==0)
      dct_diff-= (1<<size) - 1;
  }
  
  return dct_diff;
}


int vld_Chroma_DC_dct_diff(DOLProcess *p)
{
  int code, size, dct_diff;

  /* decode length */
  code = bb_Show_Bits(p, 5);

  if (code<31) {
    size = DCchromtab0[code].val;
    bb_Advance_Bits(p, DCchromtab0[code].len);
  }
  else {
    code = bb_Show_Bits(p, 10) - 0x3e0;
    size = DCchromtab1[code].val;
    bb_Advance_Bits(p, DCchromtab1[code].len);
  }

  if (size==0)
    dct_diff = 0;
  else {
    dct_diff = bb_Get_Bits(p, size);

    if ((dct_diff & (1<<(size-1)))==0)
      dct_diff-= (1<<size) - 1;
  }

  return dct_diff;
}

/* variable length decode one intra coded MPEG-1 block and
   write the coefficients to the send_buffer of p */
void vld_MPEG1_Intra_Block(DOLProcess *p, int comp)
{
  QuantizerMatrices *qm = &(p->local->quantizer_matrices);  
  MbDispatchedDatatype *mdd = &(p->local->mb_dispatched_datatype);
  int val, i, j, sign;
  unsigned int code;
  DCTtab *tab;
  int *bp;
  int *dc_dct_pred = p->local->dc_dct_predictor;

  /* check if mb_buffer big enough */
  if(p->local->send_buffer_pos > MAX_MB_BUFFER_SIZE - 64)
      Error(p, "mb_buffer to small for all blocks");
  
  /* set pointer to the begin of next block in macroblock buffer */
  bp = &(p->local->send_buffer[p->local->send_buffer_pos]);
  
  /* prepare pointer for the next block allready */
  p->local->send_buffer_pos += 64;
  
  /* ISO/IEC 11172-2 section 2.4.3.7: Block layer. */
  /* decode DC coefficients */
  if (comp<4)
    bp[0] = (dc_dct_pred[0]+=vld_Luma_DC_dct_diff(p)) << 3;
  else if (comp==4)
    bp[0] = (dc_dct_pred[1]+=vld_Chroma_DC_dct_diff(p)) << 3;
  else
    bp[0] = (dc_dct_pred[2]+=vld_Chroma_DC_dct_diff(p)) << 3;

  /* D-pictures do not contain AC coefficients */
  if(mdd->picture_coding_type == D_TYPE)
    return;

  /* decode AC coefficients */
  for (i=1; ; i++)
  {
    code = bb_Show_Bits(p, 16);
    if (code>=16384)
      tab = &DCTtabnext[(code>>12)-4];
    else if (code>=1024)
      tab = &DCTtab0[(code>>8)-4];
    else if (code>=512)
      tab = &DCTtab1[(code>>6)-8];
    else if (code>=256)
      tab = &DCTtab2[(code>>4)-16];
    else if (code>=128)
      tab = &DCTtab3[(code>>3)-16];
    else if (code>=64)
      tab = &DCTtab4[(code>>2)-16];
    else if (code>=32)
      tab = &DCTtab5[(code>>1)-16];
    else if (code>=16)
      tab = &DCTtab6[code-16];
    else
      Error(p, "invalid Huffman code in Decode_MPEG1_Intra_Block()");

    bb_Advance_Bits(p, tab->len);

    if (tab->run==64) /* end_of_block */
      return;

    if (tab->run==65) /* escape */
    {
      i+= bb_Get_Bits(p, 6);

      val = bb_Get_Bits(p, 8);
      if (val==0)
        val = bb_Get_Bits(p, 8);
      else if (val==128)
        val = bb_Get_Bits(p, 8) - 256;
      else if (val>128)
        val -= 256;

      if((sign = (val<0)))
        val = -val;
    }
    else
    {
      i+= tab->run;
      val = tab->level;
      sign = bb_Get_Bits(p, 1);
    }

    if (i>=64)
      Error(p,"DCT coeff index (i) out of bounds (intra)");

    j = scan[ZIG_ZAG][i];
    val = (val*p->local->quantizer_scale*qm->intra_quantizer_matrix[j]) >> 3;
    
    /* mismatch control ('oddification') */
    if (val!=0) /* should always be true, but it's not guaranteed */
      val = (val-1) | 1; /* equivalent to: if ((val&1)==0) val = val - 1; */

    /* saturation */
    if (!sign)
      bp[j] = (val>2047) ?  2047 :  val; /* positive */
    else
      bp[j] = (val>2048) ? -2048 : -val; /* negative */
  }
}



/* variable length decode one non-intra coded MPEG-1 block
   and write the coefficients to the send_buffer of p*/
void vld_MPEG1_Non_Intra_Block(DOLProcess *p, int comp)
{
  QuantizerMatrices *qm = &(p->local->quantizer_matrices);    
  int val, i, j, sign;
  unsigned int code;
  DCTtab *tab;
  int *bp;

  /* check if mb_buffer big enough */
  if(p->local->send_buffer_pos > MAX_MB_BUFFER_SIZE - 64)
      Error(p, "mb_buffer to small for all blocks");
  
  /* set pointer to the begin of next block of macroblock buffer */
  bp = &(p->local->send_buffer[p->local->send_buffer_pos]);
  
  /* prepare pointer for the next block allready */
  p->local->send_buffer_pos += 64;

  /* decode AC coefficients */
  for (i=0; ; i++)
  {
    code = bb_Show_Bits(p, 16);
    if (code>=16384)
    {
      if (i==0)
        tab = &DCTtabfirst[(code>>12)-4];
      else
        tab = &DCTtabnext[(code>>12)-4];
    }
    else if (code>=1024)
      tab = &DCTtab0[(code>>8)-4];
    else if (code>=512)
      tab = &DCTtab1[(code>>6)-8];
    else if (code>=256)
      tab = &DCTtab2[(code>>4)-16];
    else if (code>=128)
      tab = &DCTtab3[(code>>3)-16];
    else if (code>=64)
      tab = &DCTtab4[(code>>2)-16];
    else if (code>=32)
      tab = &DCTtab5[(code>>1)-16];
    else if (code>=16)
      tab = &DCTtab6[code-16];
    else
    {
      Error(p, "invalid Huffman code in Decode_MPEG1_Non_Intra_Block()");
    }

    bb_Advance_Bits(p, tab->len);

    if (tab->run==64) /* end_of_block */
      return;

    if (tab->run==65) /* escape */
    {
      i+= bb_Get_Bits(p, 6);

      val = bb_Get_Bits(p, 8);
      if (val==0)
        val = bb_Get_Bits(p, 8);
      else if (val==128)
        val = bb_Get_Bits(p, 8) - 256;
      else if (val>128)
        val -= 256;

      if((sign = (val<0)))
        val = -val;
    }
    else
    {
      i+= tab->run;
      val = tab->level;
      sign = bb_Get_Bits(p, 1);
    }

    if (i>=64)
    {
      Error(p,"DCT coeff index (i) out of bounds (inter)");
    }

    j = scan[ZIG_ZAG][i];
    val = (((val<<1)+1)*p->local->quantizer_scale*qm->non_intra_quantizer_matrix[j]) >> 4;

    /* mismatch control ('oddification') */
    if (val!=0) /* should always be true, but it's not guaranteed */
      val = (val-1) | 1; /* equivalent to: if ((val&1)==0) val = val - 1; */

    /* saturation */
    if (!sign)
      bp[j] = (val>2047) ?  2047 :  val; /* positive */
    else
      bp[j] = (val>2048) ? -2048 : -val; /* negative */
  }
}



/* variable length decode one intra coded MPEG-2 block and
   write the coefficients to the send_buffer of p */
void vld_MPEG2_Intra_Block(DOLProcess *p, int comp)
{
  GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
  QuantizerMatrices *qm = &(p->local->quantizer_matrices);    
  int val, i, j, sign, cc, run;
  unsigned int code;
  DCTtab *tab;
  int *bp;
  unsigned char *qmat;
  int *dc_dct_pred = p->local->dc_dct_predictor;

  /* check if mb_buffer big enough */
  if(p->local->send_buffer_pos > MAX_MB_BUFFER_SIZE - 64)
      Error(p, "mb_buffer to small for all blocks");
  
  /* set pointer to the begin of next block of macroblock buffer */
  bp = &(p->local->send_buffer[p->local->send_buffer_pos]);
  
  /* prepare pointer for the next block allready */
  p->local->send_buffer_pos += 64;

  cc = (comp<4) ? 0 : (comp&1)+1;

  qmat = (comp<4 || gdd->chroma_format==CHROMA420)
         ? qm->intra_quantizer_matrix
         : qm->chroma_intra_quantizer_matrix;

  /* ISO/IEC 13818-2 section 7.2.1: decode DC coefficients */
  if (cc==0)
    val = (dc_dct_pred[0]+= vld_Luma_DC_dct_diff(p));
  else if (cc==1)
    val = (dc_dct_pred[1]+= vld_Chroma_DC_dct_diff(p));
  else
    val = (dc_dct_pred[2]+= vld_Chroma_DC_dct_diff(p));

  bp[0] = val << (3 - p->local->intra_dc_precision);

  /* decode AC coefficients */
  for (i=1; ; i++)
  {
    code = bb_Show_Bits(p, 16);
    if (code>=16384 && !(p->local->intra_vlc_format))
      tab = &DCTtabnext[(code>>12)-4];
    else if (code>=1024)
    {
      if (p->local->intra_vlc_format)
        tab = &DCTtab0a[(code>>8)-4];
      else
        tab = &DCTtab0[(code>>8)-4];
    }
    else if (code>=512)
    {
      if (p->local->intra_vlc_format)
        tab = &DCTtab1a[(code>>6)-8];
      else
        tab = &DCTtab1[(code>>6)-8];
    }
    else if (code>=256)
      tab = &DCTtab2[(code>>4)-16];
    else if (code>=128)
      tab = &DCTtab3[(code>>3)-16];
    else if (code>=64)
      tab = &DCTtab4[(code>>2)-16];
    else if (code>=32)
      tab = &DCTtab5[(code>>1)-16];
    else if (code>=16)
      tab = &DCTtab6[code-16];
    else
    {
      Error(p, "invalid Huffman code in Decode_MPEG2_Intra_Block()");
    }

    bb_Advance_Bits(p, tab->len);

    if (tab->run==64) /* end_of_block */
    {
      return;
    }

    if (tab->run==65) /* escape */
    {
      i+= run = bb_Get_Bits(p, 6);

      val = bb_Get_Bits(p, 12);
      if ((val&2047)==0)
      {
        Error(p, "invalid escape in Decode_MPEG2_Intra_Block()");
      }
      if((sign = (val>=2048)))
        val = 4096 - val;
    }
    else
    {
      i+= run = tab->run;
      val = tab->level;
      sign = bb_Get_Bits(p, 1);
    }

    if (i>=64)
    {
      Error(p, "DCT coeff index (i) out of bounds (intra2)");
    }

    j = scan[p->local->alternate_scan][i];
    val = (val * p->local->quantizer_scale * qmat[j]) >> 4;
    bp[j] = sign ? -val : val;
  }

}


/* variable length decode one non-intra coded MPEG-2 block and
   write the coefficients to the send_buffer of p */
void vld_MPEG2_Non_Intra_Block(DOLProcess *p, int comp)
{
  GopDispatchedDatatype *gdd = &(p->local->gop_dispatched_datatype);
  QuantizerMatrices *qm = &(p->local->quantizer_matrices);    
  int val, i, j, sign, run;
  unsigned int code;
  DCTtab *tab;
  int *bp;
  unsigned char *qmat;

  /* check if mb_buffer big enough */
  if(p->local->send_buffer_pos > MAX_MB_BUFFER_SIZE - 64)
      Error(p, "mb_buffer to small for all blocks");
  
  /* set pointer to the begin of next block of macroblock buffer */
  bp = &(p->local->send_buffer[p->local->send_buffer_pos]);
  
  /* prepare pointer for the next block allready */
  p->local->send_buffer_pos += 64;

  qmat = (comp<4 || gdd->chroma_format==CHROMA420)
         ? qm->non_intra_quantizer_matrix
         : qm->chroma_non_intra_quantizer_matrix;

  /* decode AC coefficients */
  for (i=0; ; i++)
  {
    code = bb_Show_Bits(p, 16);
    if (code>=16384)
    {
      if (i==0)
        tab = &DCTtabfirst[(code>>12)-4];
      else
        tab = &DCTtabnext[(code>>12)-4];
    }
    else if (code>=1024)
      tab = &DCTtab0[(code>>8)-4];
    else if (code>=512)
      tab = &DCTtab1[(code>>6)-8];
    else if (code>=256)
      tab = &DCTtab2[(code>>4)-16];
    else if (code>=128)
      tab = &DCTtab3[(code>>3)-16];
    else if (code>=64)
      tab = &DCTtab4[(code>>2)-16];
    else if (code>=32)
      tab = &DCTtab5[(code>>1)-16];
    else if (code>=16)
      tab = &DCTtab6[code-16];
    else
    {
      Error(p, "invalid Huffman code in Decode_MPEG2_Non_Intra_Block()");
    }

    bb_Advance_Bits(p, tab->len);

    if (tab->run==64) /* end_of_block */
    {
      return;
    }

    if (tab->run==65) /* escape */
    {
      i+= run = bb_Get_Bits(p, 6);

      val = bb_Get_Bits(p, 12);
      if ((val&2047)==0)
      {
        Error(p, "invalid escape in Decode_MPEG2_Intra_Block()");
      }
      if((sign = (val>=2048)))
        val = 4096 - val;
    }
    else
    {
      i+= run = tab->run;
      val = tab->level;
      sign = bb_Get_Bits(p, 1);
    }

    if (i>=64)
    {
      Error(p,"DCT coeff index (i) out of bounds (inter2)");
    }

    j = scan[p->local->alternate_scan][i];
    val = (((val<<1)+1) * p->local->quantizer_scale * qmat[j]) >> 5;
    bp[j] = sign ? -val : val;

  }
}

